import mth5.mth5
from aurora.config import BANDS_DEFAULT_FILE
from aurora.config.config_creator import ConfigCreator

from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.sandbox.mth5_channel_summary_helpers import (
    channel_summary_to_make_mth5,
)
from aurora.test_utils.parkfield.make_parkfield_mth5 import ensure_h5_exists
from aurora.test_utils.parkfield.path_helpers import AURORA_RESULTS_PATH
from aurora.test_utils.parkfield.path_helpers import EMTF_RESULTS_PATH
from aurora.transfer_function.kernel_dataset import KernelDataset
from aurora.transfer_function.plot.comparison_plots import compare_two_z_files

from mth5.mth5 import MTH5
from mth5.helpers import close_open_files


def test_stuff_that_belongs_elsewhere():
    """
    ping the mth5, extract the summary and pass it to channel_summary_to_make_mth5

    This test was created so that codecov would see channel_summary_to_make_mth5().
    ToDo: channel_summary_to_make_mth5() method should be moved into mth5 and removed
    from aurora, including this test.

    Returns
    -------

    """
    close_open_files()
    h5_path = ensure_h5_exists()

    mth5_obj = mth5.mth5.MTH5()
    mth5_obj = MTH5(file_version="0.1.0")
    mth5_obj.open_mth5(h5_path, mode="a")
    df = mth5_obj.channel_summary.to_dataframe()
    make_mth5_df = channel_summary_to_make_mth5(df, network="NCEDC")
    mth5_obj.close_mth5()
    return make_mth5_df


def test_processing(z_file_path=None):
    """
    Parameters
    ----------
    z_file_path: str or Path or None
        Where to store zfile

    Returns
    -------
    tf_cls: TF object mt_metadata.transfer_functions.core.TF
    """

    close_open_files()
    h5_path = ensure_h5_exists()
    h5s_list = [
        h5_path,
    ]
    run_summary = RunSummary()
    run_summary.from_mth5s(h5s_list)
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary, "PKD", "SAO")

    cc = ConfigCreator()
    config = cc.create_from_kernel_dataset(
        tfk_dataset,
        output_channels=["ex", "ey"],
    )

    show_plot = False
    tf_cls = process_mth5(
        config,
        tfk_dataset,
        units="MT",
        show_plot=show_plot,
        z_file_path=z_file_path,
    )

    # tf_cls.write(fn="emtfxml_test.xml", file_type="emtfxml")
    return tf_cls


def test():

    import logging
    from mth5.helpers import close_open_files

    logging.getLogger("matplotlib.font_manager").disabled = True
    logging.getLogger("matplotlib.ticker").disabled = True

    test_stuff_that_belongs_elsewhere()
    z_file_path = AURORA_RESULTS_PATH.joinpath("pkd.zrr")
    test_processing(z_file_path=z_file_path)

    # COMPARE WITH ARCHIVED Z-FILE
    auxilliary_z_file = EMTF_RESULTS_PATH.joinpath("PKD_272_00.zrr")
    if z_file_path.exists():
        compare_two_z_files(
            z_file_path,
            auxilliary_z_file,
            label1="aurora",
            label2="emtf",
            scale_factor1=1,
            out_file="RR.png",
            markersize=3,
            rho_ylims=[1e0, 1e3],
            xlims=[0.05, 500],
        )
    else:
        print("Z-File not found - Parkfield tests failed to generate output")
        print("NCEDC probably not returning data")
    close_open_files()


if __name__ == "__main__":
    test()



from mt_metadata.timeseries.location import Location
from mth5.mth5 import MTH5


def test_can_add_location():
    """
    20220624: This mini test is being factored out of the normal tests.
    Want to assign location.  The location lands in the station_group but not in the
    run_group.
    Reported this conundrum https://github.com/kujaku11/mt_metadata/issues/85

    When I close and reopen the file, I cannot find the latitude I set anywhere.

    Returns
    -------

    """
    # make an MTH5
    m = MTH5(file_version="0.1.0")
    m.open_mth5("location_test.h5", mode="w")
    station_group = m.add_station("eureka")

    location = Location()
    location.latitude = 17.996
    station_group.metadata.location = location
    # setting latitude as above does not wind up in run, but is in station_group

    run_group = station_group.add_run("001")
    run_group.station_group.metadata.location = location
    # setting latitude as above line does not wind up in the run either"

    print("Why don't the following values agree??")
    print(f"station group {station_group.metadata.location.latitude}")
    print(f"Run Group {run_group.station_group.metadata.location.latitude}")
    m.close_mth5()

    print("Reopen the file and check if update was done on close()")
    m.open_mth5("location_test.h5", mode="r")
    eureka = m.get_station("eureka")
    print(f"station group {eureka.metadata.location.latitude}")
    run_001 = eureka.get_run("001")
    print(f"Run Group {run_001.station_group.metadata.location.latitude}")
    m.close_mth5()
    return


def main():
    test_can_add_location()


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
import psutil


from aurora.pipelines.helpers import initialize_config
from aurora.pipelines.time_series_helpers import prototype_decimate
from aurora.transfer_function.kernel_dataset import KernelDataset
from mth5.utils.exceptions import MTH5Error
from mth5.utils.helpers import initialize_mth5
from mt_metadata.transfer_functions.core import TF
from mt_metadata.transfer_functions.processing.aurora import Processing


class TransferFunctionKernel(object):
    def __init__(self, dataset=None, config=None):
        """

        Parameters
        ----------
        dataset: aurora.transfer_function.kernel_dataset.KernelDataset
        config: aurora.config.metadata.processing.Processing
        """
        processing_config = initialize_config(config)
        self._config = processing_config
        self._dataset = dataset
        self._mth5_objs = None

    @property
    def dataset(self):
        return self._dataset

    @property
    def kernel_dataset(self):
        return self._dataset

    @property
    def dataset_df(self):
        return self._dataset.df

    @property
    def processing_config(self):
        return self._config

    @property
    def config(self):
        return self._config

    @property
    def processing_summary(self):
        if self._processing_summary is None:
            self.make_processing_summary()
        return self._processing_summary

    @property
    def mth5_objs(self):
        if self._mth5_objs is None:
            self.initialize_mth5s()
        return self._mth5_objs

    def initialize_mth5s(self, mode="r"):
        """
        returns a dict of open mth5 objects, keyed by station_id

        A future version of this for multiple station processing may need nested dict with [survey_id][station_id]

        Returns
        -------
        mth5_objs : dict
            Keyed by station_ids.
            local station id : mth5.mth5.MTH5
            remote station id: mth5.mth5.MTH5
        """

        local_mth5_obj = initialize_mth5(
            self.config.stations.local.mth5_path, mode=mode
        )
        if self.config.stations.remote:
            remote_path = self.config.stations.remote[0].mth5_path
            remote_mth5_obj = initialize_mth5(remote_path, mode="r")
        else:
            remote_mth5_obj = None

        mth5_objs = {self.config.stations.local.id: local_mth5_obj}
        if self.config.stations.remote:
            mth5_objs[self.config.stations.remote[0].id] = remote_mth5_obj
        self._mth5_objs = mth5_objs
        return

    def update_dataset_df(self, i_dec_level):
        """
        This function has two different modes.  The first mode initializes values in the
        array, and could be placed into TFKDataset.initialize_time_series_data()
        The second mode, decimates. The function is kept in pipelines becasue it calls
        time series operations.

        Notes:
        1. When assigning xarrays to dataframe cells, df dislikes xr.Dataset,
        so we convert to DataArray before assignment


        Parameters
        ----------
        i_dec_level: int
            decimation level id, indexed from zero
        config: mt_metadata.transfer_functions.processing.aurora.decimation_level.DecimationLevel
            decimation level config

        Returns
        -------
        dataset_df: pd.DataFrame
            Same df that was input to the function but now has columns:


        """
        if i_dec_level == 0:
            # ANY MERGING OF RUNS IN TIME DOMAIN WOULD GO HERE

            # Assign additional columns to dataset_df, populate with mth5_objs and xr_ts
            self.dataset.initialize_dataframe_for_processing(self.mth5_objs)

            # APPLY TIMING CORRECTIONS HERE
        else:
            print(f"DECIMATION LEVEL {i_dec_level}")

            for i, row in self.dataset_df.iterrows():
                if not self.is_valid_dataset(row, i_dec_level):
                    continue
                if row.fc:
                    row_ssr_str = f"survey: {row.survey}, station_id: {row.station_id}, run_id: {row.run_id}"
                    msg = f"FC already exists for {row_ssr_str} -- skipping decimation"
                    print(msg)
                    continue
                run_xrds = row["run_dataarray"].to_dataset("channel")
                decimation = self.config.decimations[i_dec_level].decimation
                decimated_xrds = prototype_decimate(decimation, run_xrds)
                self.dataset_df["run_dataarray"].at[i] = decimated_xrds.to_array(
                    "channel"
                )  # See Note 1 above

        print("DATASET DF UPDATED")
        return

    def apply_clock_zero(self, dec_level_config):
        """get clock-zero from data if needed"""
        if dec_level_config.window.clock_zero_type == "data start":
            dec_level_config.window.clock_zero = str(self.dataset_df.start.min())
        return dec_level_config

    def check_if_fc_levels_already_exist(self):
        """
        Iterate over the processing summary_df, grouping by unique "Survey-Station-Run"s.
        (Could also iterate over kernel_dataset.dataframe, to get the groupby).

        If all FC Levels for a given station-run are already built, mark the RunSummary with a True in
        the "fc" column.  Otherwise its False

        Note 1:  Because decimation is a cascading operation, we avoid the case where some (valid) decimation
        levels exist in the mth5 FC archive and others do not.  The maximum granularity tolerated will be at the
        "station-run level, so for a given run, either all relevant FCs are packed into the h5 or we treat as if none
        of them are.  Sounds harsh, but if you want to add the logic otherwise, feel free.  If one wanted to support
        variations at the decimation-level, an appropriate way to address would be to store teh decimated time series
        in the archive as well (they would simply be runs with different sample rates, and some extra filters).

        Note 2: At this point in the logic, it is established that there are FCs associated with run_id and there are
        at least as many FC decimation levels as we require as per the processing config.  The next step is to
        assert whether it is True that the existing FCs conform to the recipe in the processing config.

        Note #3: Need to update mth5_objs dict so that it is keyed by survey, then station, else might break when
        mixing in data from other surveys (if the stations are named the same.  This can be addressed in
        the initialize_mth5s() method of TFK.  When addresssing the above issue -- consider adding the mth5_obj to
         self.dataset_df instead of keeping the dict around ..., the concern about doing this is that multiple rows
         of the dataset_df may refernece the same h5, and I don't know if updating one row will have unintended
         consequences.

        Note #4: associated_run_sub_df may have multiple rows, even though the run id is unique.
        This could happen for example when you have a long run at the local station, but multiple (say two) shorter runs
        at the reference station.  In that case, the processing summary will have a separate row for the
        intersection of the long run with each of the remote runs. We ignore this for now, selecting only the first
        element of the associated_run_sub_df, under the assumption that FCs have been created for the entire run,
        or not at all.  This assumption can be relaxed in future by using the time_period attribute of the FC layer.
        For now, we proceed with the all-or-none logic.  That is, if a ['survey', 'station_id', 'run_id',] has FCs,
        assume that the FCs are present for the entire run. We assign the "fc" column of dataset_df to have the same
         boolean value for all rows of same  ['survey', 'station_id', 'run_id',] .

        Returns: None
            Modifies self.dataset_df inplace, assigning bools to the "fc" column

        """
        groupby = [
            "survey",
            "station_id",
            "run_id",
        ]
        grouper = self.processing_summary.groupby(groupby)

        for (survey_id, station_id, run_id), df in grouper:
            cond1 = self.dataset_df.survey == survey_id
            cond2 = self.dataset_df.station_id == station_id
            cond3 = self.dataset_df.run_id == run_id
            associated_run_sub_df = self.dataset_df[cond1 & cond2 & cond3]

            if len(associated_run_sub_df) > 1:
                # See Note #4
                print(
                    "Warning -- not all runs will processed as a continuous chunk -- in future may need to loop over runlets to check for FCs"
                )

            dataset_df_indices = np.r_[associated_run_sub_df.index]
            # dataset_df_indices = associated_run_sub_df.index.to_numpy()
            run_row = associated_run_sub_df.iloc[0]
            row_ssr_str = f"survey: {run_row.survey}, station_id: {run_row.station_id}, run_id: {run_row.run_id}"

            # See Note #3 above relating to mixing multiple surveys in a processing scheme
            mth5_obj = self.mth5_objs[station_id]
            survey_obj = mth5_obj.get_survey(survey_id)
            station_obj = survey_obj.stations_group.get_station(station_id)
            if not station_obj.fourier_coefficients_group.groups_list:
                msg = f"Prebuilt Fourier Coefficients not detected for {row_ssr_str} -- will need to build them "
                print(msg)
                self.dataset_df.loc[dataset_df_indices, "fc"] = False
            else:
                print(
                    "Prebuilt Fourier Coefficients detected -- checking if they satisfy processing requirements..."
                )
                # Assume FC Groups are keyed by run_id, check if there is a relevant group
                try:
                    fc_group = station_obj.fourier_coefficients_group.get_fc_group(
                        run_id
                    )
                except MTH5Error:
                    self.dataset_df.loc[dataset_df_indices, "fc"] = False
                    print(
                        f"Run ID {run_id} not found in FC Groups, -- will need to build them "
                    )
                    continue

                if (
                    len(fc_group.groups_list)
                    < self.processing_config.num_decimation_levels
                ):
                    self.dataset_df.loc[dataset_df_indices, "fc"] = False
                    print(
                        f"Not enough FC Groups available for {row_ssr_str} -- will need to build them "
                    )
                    continue

                # Can check time periods here if desired, but unique (survey, station, run) should make this unneeded
                # processing_run = self.processing_config.stations.local.get_run(run_id)
                # for tp in processing_run.time_periods:
                #    assert tp in fc_group time periods

                # See note #2
                fcs_already_there = fc_group.supports_aurora_processing_config(
                    self.processing_config, run_row.remote
                )
                self.dataset_df.loc[dataset_df_indices, "fc"] = fcs_already_there

        return

    def make_processing_summary(self):
        """
        Melt the decimation levels over the run summary.  Add columns to estimate
        the number of FFT windows for each row

        Returns
        -------
        processing_summary: pd.DataFrame
            One row per each run-deciamtion pair
        """
        from aurora.time_series.windowing_scheme import WindowingScheme

        # Create correctly shaped dataframe
        tmp = self.kernel_dataset.df.copy(deep=True)
        id_vars = list(tmp.columns)
        decimation_info = self.config.decimation_info()
        for i_dec, dec_factor in decimation_info.items():
            tmp[i_dec] = dec_factor
        tmp = tmp.melt(id_vars=id_vars, value_name="dec_factor", var_name="dec_level")
        sortby = ["survey", "station_id", "run_id", "start", "dec_level"]
        tmp.sort_values(by=sortby, inplace=True)
        tmp.reset_index(drop=True, inplace=True)
        tmp.drop("sample_rate", axis=1, inplace=True)  # not valid for decimated data

        # Add window info
        group_by = [
            "survey",
            "station_id",
            "run_id",
            "start",
        ]
        groups = []
        grouper = tmp.groupby(group_by)
        for group, df in grouper:
            print(group)
            print(df)
            try:
                try:
                    assert (
                        df.dec_level.diff()[1:] == 1
                    ).all()  # dec levels increment by 1
                except AssertionError:
                    print(f"Skipping {group} because decimation levels are messy.")
                    continue
                assert df.dec_factor.iloc[0] == 1
                assert df.dec_level.iloc[0] == 0
            except AssertionError:
                raise AssertionError("Decimation levels not structured as expected")
            # df.sample_rate /= np.cumprod(df.dec_factor)  # better to take from config
            window_params_df = self.config.window_scheme(as_type="df")
            df.reset_index(inplace=True, drop=True)
            df = df.join(window_params_df)
            df["num_samples"] = np.floor(df.duration * df.sample_rate)
            num_windows = np.zeros(len(df))
            for i, row in df.iterrows():
                ws = WindowingScheme(
                    num_samples_window=row.num_samples_window,
                    num_samples_overlap=row.num_samples_overlap,
                )
                num_windows[i] = ws.available_number_of_windows(row.num_samples)
            df["num_stft_windows"] = num_windows
            groups.append(df)

            # tmp.loc[df.index, "sample_rate"] = df.sample_rate
        processing_summary = pd.concat(groups)
        processing_summary.reset_index(drop=True, inplace=True)
        print(processing_summary)
        self._processing_summary = processing_summary
        return processing_summary

    def validate_decimation_scheme_and_dataset_compatability(
        self, min_num_stft_windows=None
    ):
        """
        Refers to issue #182 (and #103, and possibly #196 and #233).
        Determine if there exist (one or more) runs that will yield decimated datasets
        that have too few samples to be passed to the STFT algorithm.

        Strategy for handling this:
        Mark as invlaid any rows of the processing summary that do not yield long
        enough time series to window.  This way all other rows, with decimations up to
        the invalid cases will still process.


        WCGW: If the runs are "chunked" we could encounter a situation where the
        chunk fails to yield a deep decimation level, yet the level could be produced
        if the chunk size were larger.
        In general, handling this seems a bit complicated.  We will ignore this for
        now and assume that the user will select a chunk size that is appropriate to
        the decimation scheme, i.e. use large chunks for deep decimations.

        A general solution: return a log that tells the user about each run and
        decimation level ... how many STFT-windows it yielded at each decimation level.
        This conjures the notion of (run, decimation_level) pairs
        -------

        """
        if min_num_stft_windows is None:
            min_stft_window_info = {
                x.decimation.level: x.min_num_stft_windows
                for x in self.processing_config.decimations
            }
            min_stft_window_list = [
                min_stft_window_info[x] for x in self.processing_summary.dec_level
            ]
            min_num_stft_windows = pd.Series(min_stft_window_list)

        self.processing_summary["valid"] = (
            self.processing_summary.num_stft_windows >= min_num_stft_windows
        )

    def validate_processing(self):
        """
        Things that are validated:
        1. The default estimation engine from the json file is "RME_RR", which is fine (
        we expect to in general to do more RR processing than SS) but if there is only
        one station (no remote)then the RME_RR should be replaced by default with "RME".

        2. make sure local station id is defined (correctly from kernel dataset)
        """

        # Make sure a RR method is not being called for a SS config
        if not self.config.stations.remote:
            for decimation in self.config.decimations:
                if decimation.estimator.engine == "RME_RR":
                    print("No RR station specified, switching RME_RR to RME")
                    decimation.estimator.engine = "RME"

        # Make sure that a local station is defined
        if not self.config.stations.local.id:
            print("WARNING: Local station not specified")
            print("Setting local station from Kernel Dataset")
            self.config.stations.from_dataset_dataframe(self.kernel_dataset.df)

    def validate(self):
        self.validate_processing()
        self.validate_decimation_scheme_and_dataset_compatability()
        if self.memory_warning():
            raise Exception

    def valid_decimations(self):
        """

        Returns
        -------

        """
        # identify valid rows of processing summary
        tmp = self.processing_summary[self.processing_summary.valid]
        valid_levels = tmp.dec_level.unique()

        dec_levels = [x for x in self.config.decimations]
        dec_levels = [x for x in dec_levels if x.decimation.level in valid_levels]
        print(f"After validation there are {len(dec_levels)} valid decimation levels")
        return dec_levels

    def is_valid_dataset(self, row, i_dec):
        """
        Given a row from the RunSummary, answers:
        Will this decimation level yield a valid dataset?

        Parameters
        ----------
        row: pandas.core.series.Series
            Row of the self._dataset_df (corresponding to a run that will be processed)
        i_dec: integer
            refers to decimation level

        Returns
        -------
        is_valid: Bool
            Whether the (run, decimation_level) pair associated with this row yields
            a valid dataset
        """
        # Query processing on survey, station, run, decimation
        cond1 = self.processing_summary.survey == row.survey
        cond2 = self.processing_summary.station_id == row.station_id
        cond3 = self.processing_summary.run_id == row.run_id
        cond4 = self.processing_summary.dec_level == i_dec
        cond5 = self.processing_summary.start == row.start

        cond = cond1 & cond2 & cond3 & cond4 & cond5
        processing_row = self.processing_summary[cond]
        if len(processing_row) != 1:
            return False
        is_valid = processing_row.valid.iloc[0]
        return is_valid

    @property
    def processing_type(self):
        """
        A description of the processing, will get passed to TF object,
        can be used for Z-file

        Could add a version or a hashtag to this
        Could also check dataset_df
        If remote.all==False append "Single Station"
        """
        processing_type = "Aurora"
        if self.dataset_df.remote.any():
            processing_type = f"{processing_type} Robust Remote Reference"
        else:
            processing_type = f"{processing_type} Robust Single Station"
        if "processing_type" in self.dataset_df.columns:
            processing_type = self.dataset_df["processing_type"].iloc[0]

        return processing_type

    def export_tf_collection(self, tf_collection):
        """
        Assign transfer_function, residual_covariance, inverse_signal_power, station, survey

        Parameters
        ----------
        tf_collection: aurora.transfer_function.transfer_function_collection.TransferFunctionCollection
            Contains TF estimates, covariance, and signal power values

        Returns
        -------
        tf_cls: mt_metadata.transfer_functions.core.TF
            Transfer function container
        """

        def make_decimation_dict_for_tf(tf_collection, processing_config):
            """
            Decimation dict is used by mt_metadata's TF class when it is writng z-files.
            If no z-files will be written this is not needed

            sample element of decimation_dict:
            '1514.70134': {'level': 4, 'bands': (5, 6), 'npts': 386, 'df': 0.015625}}

            Note #1:  The line that does
            period_value["npts"] = tf_collection.tf_dict[i_dec].num_segments.data[0, i_band]
            doesn't feel very robust ... it would be better to explicitly select based on the num_segments
            xarray location where period==fb.center_period.  This is a placeholder for now.  In future,
            the TTFZ class is likely to be deprecated in favour of directly packing mt_metadata TFs during processing.

            Parameters
            ----------
            tfc

            Returns
            -------

            """
            from mt_metadata.transfer_functions.io.zfiles.zmm import PERIOD_FORMAT

            decimation_dict = {}

            for i_dec, dec_level_cfg in enumerate(processing_config.decimations):
                for i_band, band in enumerate(dec_level_cfg.bands):
                    period_key = f"{band.center_period:{PERIOD_FORMAT}}"
                    period_value = {}
                    period_value["level"] = i_dec + 1  # +1 to match EMTF standard
                    period_value["bands"] = tuple(band.harmonic_indices[np.r_[0, -1]])
                    period_value["sample_rate"] = dec_level_cfg.sample_rate_decimation
                    try:
                        period_value["npts"] = tf_collection.tf_dict[
                            i_dec
                        ].num_segments.data[0, i_band]
                    except KeyError:
                        print("Possibly invalid decimation level")
                        period_value["npts"] = 0
                    decimation_dict[period_key] = period_value

            return decimation_dict

        channel_nomenclature = self.config.channel_nomenclature
        channel_nomenclature_dict = channel_nomenclature.to_dict()[
            "channel_nomenclature"
        ]
        merged_tf_dict = tf_collection.get_merged_dict(channel_nomenclature)
        decimation_dict = make_decimation_dict_for_tf(
            tf_collection, self.processing_config
        )

        tf_cls = TF(
            channel_nomenclature=channel_nomenclature_dict,
            decimation_dict=decimation_dict,
        )
        renamer_dict = {"output_channel": "output", "input_channel": "input"}
        tmp = merged_tf_dict["tf"].rename(renamer_dict)
        tf_cls.transfer_function = tmp

        isp = merged_tf_dict["cov_ss_inv"]
        renamer_dict = {"input_channel_1": "input", "input_channel_2": "output"}
        isp = isp.rename(renamer_dict)
        tf_cls.inverse_signal_power = isp

        res_cov = merged_tf_dict["cov_nn"]
        renamer_dict = {"output_channel_1": "input", "output_channel_2": "output"}
        res_cov = res_cov.rename(renamer_dict)
        tf_cls.residual_covariance = res_cov

        # Set key as first el't of dict, nor currently supporting mixed surveys in TF
        tf_cls.survey_metadata = self.dataset.local_survey_metadata
        tf_cls.station_metadata.transfer_function.processing_type = self.processing_type
        return tf_cls

    def memory_warning(self):
        """
        Checks if we should be anitcipating a RAM issue
        Requires an estimate of available RAM, and an estimate of the dataset size
        Available RAM is taken from psutil,
        Dataset size is number of samples, times the number of bytes per sample
        Bits per sample is estimated to be 64 by default, which is 8-bytes
        Returns
        -------

        """
        bytes_per_sample = 8  # 64-bit
        memory_threshold = 0.5

        # get the total amount of RAM in bytes
        total_memory = psutil.virtual_memory().total

        # print the total amount of RAM in GB
        print(f"Total memory: {total_memory / (1024 ** 3):.2f} GB")
        num_samples = self.dataset_df.duration * self.dataset_df.sample_rate
        total_samples = num_samples.sum()
        total_bytes = total_samples * bytes_per_sample
        print(f"Total Bytes of Raw Data: {total_bytes / (1024 ** 3):.3f} GB")

        ram_fraction = 1.0 * total_bytes / total_memory
        print(f"Raw Data will use: {100 * ram_fraction:.3f} % of memory")

        # Check a condition
        if total_bytes > memory_threshold * total_memory:
            return True
        else:
            return False



import subprocess
import unittest
import os
import platform


class Doc_Test(unittest.TestCase):
    @property
    def path_to_docs(self):
        dirname, file_name = os.path.split(os.path.abspath(__file__))
        return dirname.split(os.path.sep)[:-1] + ["docs"]

    def test_html(self):
        wd = os.getcwd()
        os.chdir(os.path.sep.join(self.path_to_docs))

        if platform.system() != "Windows":
            response = subprocess.run(["make", "html"])
            self.assertTrue(response.returncode == 0)
        else:
            response = subprocess.call(["make", "html"], shell=True)
            self.assertTrue(response == 0)

        os.chdir(wd)


if __name__ == "__main__":
    unittest.main()



"""


"""
import scipy.signal as ssig
from mt_metadata.transfer_functions.processing.aurora.decimation_level import (
    get_fft_harmonics,
)


def butter_bandpass(low_cut, high_cut, sample_rate, order=5):
    """
    https://scipy.github.io/old-wiki/pages/Cookbook/ButterworthBandpass
    """
    nyquist_frequency = 0.5 * sample_rate
    low = low_cut / nyquist_frequency
    high = high_cut / nyquist_frequency
    b, a = ssig.butter(order, [low, high], btype="band", analog=False)
    return b, a


def butter_highpass(f_cut, fs, order=5):
    """
    This is based on butter_bandpass
    """
    f_nyq = 0.5 * fs
    normalized_frequency = f_cut / f_nyq
    b, a = ssig.butter(order, normalized_frequency, btype="high", analog=False)
    return b, a


def butter_lowpass(cutoff, fs, order=5):
    """
    This is based on butter_bandpass
    """
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = ssig.butter(order, normal_cutoff, btype="low", analog=False)
    return b, a


def butter_highpass_by_harmonic_index(
    num_samples=128,
    sample_rate=1.0,
    n_harmonic_cutoff=3,
    filter_order=11,
):
    """
    application of the sos filter created could use for example:
    for ch in local["mvts"].keys():
        local["mvts"][ch].data = ssig.sosfilt(sos, local["mvts"][ch].data)

    Parameters
    ----------
    num_samples_window
    sample_rate
    n_harmonic_cutoff
    filter_order

    Returns
    -------

    """
    freqs = get_fft_harmonics(num_samples, sample_rate)
    sos = ssig.butter(
        filter_order,
        freqs[n_harmonic_cutoff],
        btype="highpass",
        fs=sample_rate,
        output="sos",
    )
    return sos



from aurora.general_helper_functions import BAND_SETUP_PATH
import pathlib

BANDS_DEFAULT_FILE = BAND_SETUP_PATH.joinpath("bs_test.cfg")
BANDS_256_26_FILE = BAND_SETUP_PATH.joinpath("bs_256_26.cfg")
BANDS_256_29_FILE = BAND_SETUP_PATH.joinpath("bs_256_29.cfg")



"""
follows Gary's TTF.m in iris_mt_scratch
egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes
"""

import numpy as np
import xarray as xr

from mt_metadata.base import Base


class TransferFunction(Base):
    """
    Class to contain transfer function array.

    Parameters:
    TF : numpy array
        array of transfer functions: TF(Nout, Nin, Nperiods)
    T : numpy array
        list of periods
    cov_ss_inv : numpy array
        inverse signal power matrix.  aka Cov_SS in EMTF matlab codes
    cov_nn : numpy array
        noise covariance matrix: aka Cov_NN in EMTF matlab codes
    num_segments : integer array?
        Number of samples used to estimate TF for each band, and for each \
        output channel (might be different for different channels)
    R2 : xarray.DataArray
        multiple coherence for each output channel / band
    FullCov : boolean
        true if full covariance is provided

    properties (Dependent)
    StdErr % standard errors of TF components, same size and order as TF
    NBands
    freqs % inverse of period
    Nout
    Nin
    """

    def __init__(self, decimation_level_id, frequency_bands, **kwargs):
        """
        change 2021-07-23 to require a frequency_bands object.  We may want
        to just pass the band_edges.  I'm not a fan of forcing dependency of
        TF on the FrequencyBands class, but it will simplify writting the z_file
        interfaces.  The old TTF.py class shows an example that just accepted
        num_bands as an initialation variable.

        Parameters
        ----------
        _emtf_header : legacy header information used by Egbert's matlab class.  Header contains
        local site header, remote site header if appropriate, and information about estimation approach
        decimation_level_id: int
            Identifies the relevant decimation level.  Used for accessing the
            appropriate info in self.processing config.
        frequency_bands: aurora.time_series.frequency_band.FrequencyBands
            frequency bands object
        """
        self._emtf_tf_header = None
        self.decimation_level_id = decimation_level_id
        self.frequency_bands = frequency_bands
        self.num_segments = None
        self.cov_ss_inv = None
        self.cov_nn = None
        self.R2 = None
        self.initialized = False
        self.processing_config = kwargs.get("processing_config", None)

        if self.emtf_tf_header is not None:
            if self.num_bands is not None:
                self._initialize_arrays()

    @property
    def emtf_tf_header(self):
        if self.processing_config is None:
            print("No header is available without a processing config")
            self._emtf_tf_header = None
        else:
            if self._emtf_tf_header is None:
                tfh = self.processing_config.emtf_tf_header(self.decimation_level_id)
                self._emtf_tf_header = tfh
        return self._emtf_tf_header

    @property
    def tf_header(self):
        return self.emtf_tf_header

    @property
    def tf(self):
        return self.transfer_function

    @property
    def num_bands(self):
        """
        Returns num_bands : int
            a count of the frequency bands associated with the TF
        -------

        """
        return self.frequency_bands.number_of_bands

    @property
    def periods(self):
        periods = self.frequency_bands.band_centers(frequency_or_period="period")
        periods = np.flipud(periods)
        return periods

    def _initialize_arrays(self):
        """
        There are four separate data strucutres, indexed by period here:

        TF (num_channels_out, num_channels_in)
        cov_ss_inv (num_channels_in, num_channels_in, num_bands),
        Cov_NN (num_channels_out, num_channels_out),
        R2 num_channels_out)

        We use frequency in Fourier domain and period in TF domain.  Might
        be better to be consistent.  It would be inexpensive to duplicate
        the axis and simply provide both.

        Each of these is indexed by period (or frequency)
        TODO: These may be better cast as xarrays.  Review after curves up
        and running

        TODO: also, I prefer np.full(dim_tuple, np.nan) for init
        Returns
        -------

        """
        if self.tf_header is None:
            print("header needed to allocate transfer function arrays")
            raise Exception

        # <transfer function xarray>
        tf_array_dims = (self.num_channels_out, self.num_channels_in, self.num_bands)
        tf_array = np.zeros(tf_array_dims, dtype=np.complex128)
        self.transfer_function = xr.DataArray(
            tf_array,
            dims=["output_channel", "input_channel", "period"],  # frequency"],
            coords={
                "period": self.periods,
                "output_channel": self.tf_header.output_channels,
                "input_channel": self.tf_header.input_channels,
            },
        )

        # num_segments xarray
        num_segments = np.zeros((self.num_channels_out, self.num_bands), dtype=np.int32)
        num_segments_xr = xr.DataArray(
            num_segments,
            dims=["channel", "period"],
            coords={
                "period": self.periods,
                "channel": self.tf_header.output_channels,
            },
        )
        self.num_segments = num_segments_xr

        # Inverse signal covariance
        cov_ss_dims = (self.num_channels_in, self.num_channels_in, self.num_bands)
        cov_ss_inv = np.zeros(cov_ss_dims, dtype=np.complex128)
        self.cov_ss_inv = xr.DataArray(
            cov_ss_inv,
            dims=["input_channel_1", "input_channel_2", "period"],
            coords={
                "input_channel_1": self.tf_header.input_channels,
                "input_channel_2": self.tf_header.input_channels,
                "period": self.periods,
            },
        )

        # Noise covariance
        cov_nn_dims = (self.num_channels_out, self.num_channels_out, self.num_bands)
        cov_nn = np.zeros(cov_nn_dims, dtype=np.complex128)
        self.cov_nn = xr.DataArray(
            cov_nn,
            dims=["output_channel_1", "output_channel_2", "period"],
            coords={
                "output_channel_1": self.tf_header.output_channels,
                "output_channel_2": self.tf_header.output_channels,
                "period": self.periods,
            },
        )

        # Coefficient of determination
        self.R2 = xr.DataArray(
            np.zeros((self.num_channels_out, self.num_bands)),
            dims=["output_channel", "period"],
            coords={
                "output_channel": self.tf_header.output_channels,
                "period": self.periods,
            },
        )

        self.initialized = True

    @property
    def minimum_period(self):
        return np.min(self.periods)

    @property
    def maximum_period(self):
        return np.max(self.periods)

    @property
    def num_channels_in(self):
        return len(self.tf_header.input_channels)

    @property
    def num_channels_out(self):
        return len(self.tf_header.output_channels)

    def frequency_index(self, frequency):
        return self.period_index(1.0 / frequency)
        # frequency_index = np.isclose(self.num_segments.frequency, frequency)
        # frequency_index = np.where(frequency_index)[0][0]
        # return frequency_index

    def period_index(self, period):
        period_index = np.isclose(self.num_segments.period, period)
        period_index = np.where(period_index)[0][0]
        return period_index
        # return self.frequency_index(1.0 / period)

    def tf_to_df(self):
        pass
        # import pandas as pd
        # columns = ["input_channel", "output_channel", "frequency", ]
        #            #"decimation_level"]
        # df = pd.DataFrame(columns=columns)

    def set_tf(self, regression_estimator, period):
        """
        This sets TF elements for one band, using contents of regression_estimator
        object.  This version assumes there are estimates for Nout output channels
        """
        index = self.period_index(period)

        tf = regression_estimator.b_to_xarray()
        output_channels = list(regression_estimator._Y.data_vars)
        input_channels = list(regression_estimator._X.data_vars)

        for out_ch in output_channels:
            for inp_ch in input_channels:
                self.tf[:, :, index].loc[out_ch, inp_ch] = tf.loc[out_ch, inp_ch]

        if regression_estimator.noise_covariance is not None:
            for out_ch_1 in output_channels:
                for out_ch_2 in output_channels:
                    tmp = regression_estimator.cov_nn.loc[out_ch_1, out_ch_2]
                    self.cov_nn[:, :, index].loc[out_ch_1, out_ch_2] = tmp

        if regression_estimator.inverse_signal_covariance is not None:
            for inp_ch_1 in input_channels:
                for inp_ch_2 in input_channels:
                    tmp = regression_estimator.cov_ss_inv.loc[inp_ch_1, inp_ch_2]
                    self.cov_ss_inv[:, :, index].loc[inp_ch_1, inp_ch_2] = tmp

        if regression_estimator.R2 is not None:
            for out_ch in output_channels:
                tmp = regression_estimator.R2.loc[out_ch]
                self.R2[:, index].loc[out_ch] = tmp

        self.num_segments.data[:, index] = regression_estimator.n_data

        return



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import aurora
from sphinx_gallery.sorting import FileNameSortKey

# -- Project information -----------------------------------------------------

project = "aurora"
copyright = "2021, Karl Kappler, Jared Peacock, Andy Frassetto, Tim Ronan, Lindsey Heagy, Douglas Oldenburg"
author = "2021, Karl Kappler, Jared Peacock, Tim Ronan, Andy Frassetto, Lindsey Heagy, Douglas Oldenburg"

# The full version, including alpha/beta/rc tags
release = "0.0.1"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "matplotlib.sphinxext.plot_directive",
    "numpydoc",
    "nbsphinx",
    "sphinx_gallery.gen_gallery",
]

# Autosummary pages will be generated by sphinx-autogen instead of sphinx-build
autosummary_generate = True

numpydoc_class_members_toctree = False

# API doc options
apidoc_module_dir = "../aurora"
apidoc_output_dir = "api/generated"
apidoc_toc_file = False
apidoc_excluded_paths = []
apidoc_separate_modules = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
try:
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
    pass
except Exception:
    html_theme = "default"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
    "matplotlib": ("https://matplotlib.org/", None),
}

# Sphinx Gallery
sphinx_gallery_conf = {
    # path to your examples scripts
    "examples_dirs": [
        "../tutorials",
    ],
    "gallery_dirs": [
        "examples",
    ],
    "within_subsection_order": FileNameSortKey,
    "filename_pattern": "\.py",
    "backreferences_dir": "api/generated/backreferences",
    "doc_module": "aurora",
    # 'reference_url': {'discretize': None},
}



"""
follows Gary's IterControl.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes
"""
import numpy as np

from aurora.transfer_function.regression.helper_functions import rme_beta


class IterControl(object):
    """ """

    def __init__(
        self,
        max_number_of_iterations=10,
        max_number_of_redescending_iterations=2,
        **kwargs,
    ):
        """

        Parameters
        ----------
        max_number_of_iterations: int
            Set to zero for OLS, otherwise, this is how many times the RME
            will refine the estimate.
        max_number_of_redescending_iterations : int
            1 or 2 is fine at most.  If set to zero we ignore the redescend code block.
        tolerance : float
            minimum fractional change in any term in b.  Any smaller change
            than this will trigger convergence
        epsilon : float
            NOT USED: REMOVE
        kwargs

        Class Variables:
        <Specific to Egbert's Robust Regression>
        r0: float
            Effectively infinty for OLS, this controls the point at which residuals
            transition from being penalized by a squared vs a linear function.  The
            units are standard deviations of the residual distribution.  Normally
            1.4, 1.5 or thereabouts
        u0: float
            u0 is a parameter for the redescending portion of the robust regression.
            It is controlled by a double exponential formula (REFERENCE NEEDED).  It
            makes for severe downweighting about u0.  The function is continuous
            "math friendly" (all derivates exist etc) so you can prove theorems about it
            etc.
        </Specific to Egbert's Robust Regression>
        """
        self.number_of_iterations = 0
        self.number_of_redescending_iterations = 0

        self.tolerance = 0.005
        self.epsilon = 1000
        self.max_number_of_iterations = max_number_of_iterations
        self.max_number_of_redescending_iterations = (
            max_number_of_redescending_iterations
        )

        # <regression-M estimator params>
        self.r0 = 1.5
        self.u0 = 2.8
        # </regression-M estimator params>

        # <Additional properties>
        # used sometimes to control one or another of the iterative algorithms
        # These were translated from the matlab code and may be moved in future
        self.return_covariance = True
        self.save_cleaned = False
        self.robust_diagonalize = False
        # </Additional properties>

    def converged(self, b, b0):
        """
        Parameters
        ----------
        b : complex-valued numpy array
            the most recent regression estimate
        b0 : complex-valued numpy array
            The previous regression estimate

        Returns
        -------
        converged: bool
            True of the regression has terminated, False otherwise

        Notes:
        The variable maximum_change finds the maximum amplitude component of the vector
        1-b/b0.  Looking at the formula, one might want to cast this instead as
        1 - abs(b/b0), however, that will be insensitive to phase changes in b,
        which is complex valued.  The way it is coded np.max(np.abs(1 - b / b0)) is
        correct as it stands.
        """

        converged = False
        maximum_change = np.max(np.abs(1 - b / b0))
        tolerance_cond = maximum_change <= self.tolerance
        iteration_cond = self.number_of_iterations >= self.max_number_of_iterations
        if tolerance_cond or iteration_cond:
            converged = True
            # These print statments are not very clear and
            # Should be reworded.
            # if tolerance_cond:
            #    print(
            #        f"Converged Due to MaxChange < Tolerance after "
            #        f" {self.number_of_iterations} of "
            #        f" {self.max_number_of_iterations} iterations"
            #    )
            # elif iteration_cond:
            #    print(
            #        f"Converged Due to maximum number_of_iterations "
            #        f" {self.max_number_of_iterations}"
            #    )
        else:
            converged = False

        return converged

    @property
    def continue_redescending(self):
        maxxed_out = (
            self.number_of_redescending_iterations
            >= self.max_number_of_redescending_iterations
        )
        if maxxed_out:
            return False
        else:
            return True

    @property
    def correction_factor(self):
        """
        TODO: This is an RME specific property.  Suggest move r0, u0 and this method
        into an RME-config class.

        See notes on usage in
        transfer_function.regression.helper_functions.rme_beta

        Returns
        -------
        correction_factor : float
            correction factor used for scaling the residual error_variance
        """
        correction_factor = 1.0 / rme_beta(self.r0)
        return correction_factor



from obspy import read_inventory

from mt_metadata.timeseries.stationxml import XMLInventoryMTExperiment
from mth5.utils.helpers import initialize_mth5


def xml_to_mth5(xml_path, h5_path="tmp.h5"):
    """
    Parameters
    ----------
    xml_path

    Returns
    -------

    """
    inventory0 = read_inventory(str(xml_path))  # 8P
    translator = XMLInventoryMTExperiment()
    experiment = translator.xml_to_mt(inventory_object=inventory0)
    mth5_obj = initialize_mth5(h5_path)
    mth5_obj.from_experiment(experiment)
    return mth5_obj



"""
This module was inside of mth5/clients/helper_functions.py
on branch issue_76_make_mth5_factoring

Some of these functions are handy, and should eventually be merged into mth5.

I would also like to use some of these functions from time-to-time, so I am putting
them here for now, until we can decide what to move to mth5 and what to keep in
aurora (and what to throw out).
"""
import datetime
import pandas as pd

from mt_metadata.timeseries.stationxml import XMLInventoryMTExperiment
from mth5.clients import FDSN
from mth5.utils.helpers import initialize_mth5


def enrich_channel_summary(mth5_object, df, keyword):
    """

    Parameters
    ----------
    mth5_object: mth5.mth5.MTH5
    df: pd.DataFrame
        A channel summary dataframe
    keyword: str
        supported keywords are ["num_filters",]
        "num_filters" computes the number of filters associated with each row (channel-run) and adds that "num_filters" column of df

    Returns
    -------
    df: pd.DataFrame
        The channel summary df with the new column
    """
    df[keyword] = -1
    if keyword == "num_filters":
        for i_row, row in df.iterrows():
            channel = mth5_object.get_channel(
                row.station, row.run, row.component, row.survey
            )
            num_filters = len(channel.channel_response_filter.filters_list)
            df[keyword].iat[i_row] = num_filters
    elif keyword=="filter_units_in":
        for i_row, row in df.iterrows():
            channel = mth5_object.get_channel(row.station, row.run, row.component, row.survey)
            units_in = [x.units_in for x in channel.channel_response_filter.filters_list]
            df[keyword].iat[i_row] = units_in
    elif keyword=="filter_units_out":
        for i_row, row in df.iterrows():
            channel = mth5_object.get_channel(row.station, row.run, row.component, row.survey)
            units_out = [x.units_out for x in channel.channel_response_filter.filters_list]
            df[keyword].iat[i_row] = units_out
    return df


def augmented_channel_summary(mth5_object, df=None):  # , **kwargs):
    """
    Consider supportig kwargs, such as a list of keyords that tell what columns to add
    For now, we only want to add n_filters
    Parameters
    ----------
    df: channel summary dataframe


    Returns
    -------

    """
    if not df:
        df = mth5_object.channel_summary.to_dataframe()
    df["n_filters"] = -1
    for i_row, row in df.iterrows():
        channel = mth5_object.get_channel(
            row.station, row.run, row.component, row.survey
        )
        n_filters = len(channel.channel_response_filter.filters_list)
        df.n_filters.iat[i_row] = n_filters
    return df


def build_request_df(network_id, station_id, channels=None,
                     start=None, end=None, time_period_dict={},
                     mth5_version='0.2.0'):
    """

    Parameters
    ----------
    network_id: string
        Two-character network identifier string fro FDSN.
    station_id: string
        Short identifier code used by FDSN, e.g. CAS04, NVR11
    channels: list or None
        3-character channel identifiers, e.g. ["LQ2", "MFZ"],
        support for wildcards of the form ["*F*", "*Q*",] is experimental
        Does not support wildcards of the form ["*",]
    start: string
        ISO-8601 representation of a timestamp
    end: string
        ISO-8601 representation of a timestamp
    time_period_dict: dict
        Keyed by same values as channels, this gives explicit start/end times for each of the channels
    mth5_version: str
        From ["0.1.0", "0.2.0"]

    Returns
    -------


    Returns:
        request_df: pd.DataFrame
        A formatted dataframe that can be passed to mth5.clients.FDSN to request metdata or data.

    """
    from mth5.clients import FDSN
    def get_time_period_bounds(ch):
        if ch in time_period_dict.keys():
            time_interval = time_period_dict[ch]
            ch_start = time_period_dict[ch].left.isoformat()
            ch_end = time_period_dict[ch].right.isoformat()
        else:
            if start is None:
                ch_start = '1970-01-01 00:00:00'
            else:
                ch_start = start
            if end is None:
                ch_end = datetime.datetime.now()
                ch_end = ch_end.replace(hour=0, minute=0, second=0, microsecond=0)
                ch_end = str(ch_end)
            else:
                ch_end = end
        return ch_start, ch_end

    fdsn_object = FDSN(mth5_version=mth5_version)
    fdsn_object.client = "IRIS"

    request_list = []
    for channel in channels:
        ch_start, ch_end = get_time_period_bounds(channel)
        request_list.append([network_id, station_id, '', channel, ch_start, ch_end])

    print(f"request_list: {request_list}")

    request_df = pd.DataFrame(request_list, columns=fdsn_object.request_columns)
    # workaround for having a channel with missing run
    # request_df["start"] = request_df["start"].max()
    return request_df


def get_experiment_from_obspy_inventory(inventory):
    translator = XMLInventoryMTExperiment()
    experiment = translator.xml_to_mt(inventory_object=inventory)
    return experiment


def mth5_from_experiment(experiment, h5_path=None):
    """

    Parameters
    ----------
    experiment
    h5_path

    Returns
    -------

    """
    mth5_obj = initialize_mth5(h5_path)
    mth5_obj.from_experiment(experiment)
    return mth5_obj


def get_channel_summary(h5_path):
    """

    Parameters
    ----------
    h5_path: pathlib.Path
        Where is the h5

    Returns
    -------
    channel_summary_df: pd.DataFrame
        channel summary from mth5
    """
    mth5_obj = initialize_mth5(
        h5_path=h5_path,
    )
    mth5_obj.channel_summary.summarize()
    channel_summary_df = mth5_obj.channel_summary.to_dataframe()
    mth5_obj.close_mth5()
    print(channel_summary_df)
    return channel_summary_df



"""
Helper class to make config files.

Note: the config is still evolving and this class and its methods are expected to
change.


"""
from pathlib import Path

from aurora.config.metadata.processing import Processing
from aurora.config import BANDS_DEFAULT_FILE
from mt_metadata.transfer_functions.processing.aurora import Station, Run
from mt_metadata.transfer_functions.processing.aurora.window import Window
from aurora.sandbox.io_helpers.emtf_band_setup import EMTFBandSetupFile


class ConfigCreator:
    def __init__(self, **kwargs):
        self._band_specification_style = None

    def processing_id(self, kernel_dataset):
        """
        In the past, we used f"{local}-{remote}" or  f"{local}-{run_id}"
        Neither of these is sufficiently unique.  In fact, they only describe the
        dataset, and not the processing config.  It is difficult to see how to make a
        comprehensive, unique id without it being very long or involving hash
        functions.

        For now, will try to use {local}-{local_runs}-{remote}-{remote_runs},
        which at least describes the dataset, then a string can be generated by the
        config and appended if needed.


        Parameters
        ----------
        kernel_dataset

        Returns
        -------

        """
        id = f"{kernel_dataset.local_station_id}-{kernel_dataset.remote_station_id}"
        return id

    @property
    def band_specification_style(self):
        return self._band_specification_style

    def determine_band_specification_style(self):
        """
        ? Should emtf_band_file path be stored in config?
        It would support reproducability.

        Parameters
        ----------
        emtf_band_file
        band_edges

        """

        if (self._emtf_band_file is None) & (self._band_edges is None):
            print("Bands not defined; setting to EMTF BANDS_DEFAULT_FILE")
            self._emtf_band_file = BANDS_DEFAULT_FILE
            self._band_specification_style = "EMTF"
        elif (self._emtf_band_file is not None) & (self._band_edges is not None):
            print("Bands defined twice, and possibly inconsistantly")
            raise Exception
        elif self._band_edges is not None:
            self._band_specification_style = "band_edges"
        elif self._emtf_band_file is not None:
            self._band_specification_style = "EMTF"
        else:
            print("method of band specifciation is unrecognized")
            raise NotImplementedError

    def create_from_kernel_dataset(
        self,
        kernel_dataset,
        input_channels=["hx", "hy"],
        output_channels=["hz", "ex", "ey"],
        estimator=None,
        **kwargs,
    ):
        """
        Hmmm, why not make this a method of kernel_dataset??

        Early on we want to know how may decimation levels there will be.
        This is defined either by:
         1. decimation_factors argument (normally accompanied by a bands_dict)
         2. number of decimations implied by emtf band setup file.
        Theoretically, you could also use the number of decimations implied by
        bands_dict but this is sloppy, because it would be bad practice to assume
        the decimation factor.

        Notes:
        1.  2022-09-10
        The reading-in from EMTF band setup file used to be very terse, carried
        some baked in assumptions about decimation factors, and did not acknowlege
        specific frequency bands in Hz.  I am adding some complexity to the method
        that populates bands from EMTF band setup file but am now explict about the
        assumtion of decimation factors, and do provide the frequency bands in Hz.


        Parameters
        ----------
        kernel_dataset
        emtf_band_file: while the default here is None, it will get assigned the
        value BANDS_DEFAULT_FILE in the set_frequecy_bands method if band edges is
        also None.
        input_channels
        output_channels
        estimator
        band_edges
        kwargs

        Returns
        -------

        """

        processing_id = self.processing_id(kernel_dataset)
        processing_obj = Processing(id=processing_id)  # , **kwargs)

        # pack station and run info into processing object
        processing_obj.stations.from_dataset_dataframe(kernel_dataset.df)

        # Unpack kwargs
        self._emtf_band_file = kwargs.get("emtf_band_file", None)
        self._band_edges = kwargs.get("band_edges", None)
        decimation_factors = kwargs.get("decimation_factors", None)
        num_samples_window = kwargs.get("num_samples_window", None)

        # determine window parameters:
        # check if they have been passed as kwargs, otherwise extract default values

        # Determine if band_setup or edges dict is to be used for bands
        self.determine_band_specification_style()

        # Set Frequency Bands
        if self.band_specification_style == "EMTF":
            # see note 1
            emtf_band_setup_file = EMTFBandSetupFile(
                filepath=self._emtf_band_file, sample_rate=kernel_dataset.sample_rate
            )
            num_decimations = emtf_band_setup_file.num_decimation_levels
            if decimation_factors is None:
                # set default values to EMTF default values [1, 4, 4, 4, ..., 4]
                decimation_factors = num_decimations * [4]
                decimation_factors[0] = 1
            if num_samples_window is None:
                default_window = Window()
                num_samples_window = num_decimations * [default_window.num_samples]
            elif isinstance(num_samples_window, int):
                num_samples_window = num_decimations * [num_samples_window]
            # now you can define the frequency bands
            band_edges = emtf_band_setup_file.compute_band_edges(
                decimation_factors, num_samples_window
            )
            self._band_edges = band_edges

        processing_obj.assign_bands(
            self._band_edges,
            kernel_dataset.sample_rate,
            decimation_factors,
            num_samples_window,
        )
        processing_obj.band_specification_style = self.band_specification_style
        if self.band_specification_style == "EMTF":
            processing_obj.band_setup_file = str(self._emtf_band_file)
        for key, decimation_obj in processing_obj.decimations_dict.items():
            decimation_obj.input_channels = input_channels
            decimation_obj.output_channels = output_channels
            if num_samples_window is not None:
                decimation_obj.window.num_samples = num_samples_window[key]
            # set estimator if provided as kwarg
            if estimator:
                try:
                    decimation_obj.estimator.engine = estimator["engine"]
                except KeyError:
                    pass
        return processing_obj



"""
This script reads raw parkfield ascii data from csvs in the data repo and saves
it to an h5 file.  The flow is:

1. read in all the parkfield data to a dataframe,
2  read in all the hollister data to a dataframe,
3. Save the pkd dataframe to compressed h5
4. Save the hollister dataframe to compressed h5

This script is not intended to be run as a test.
The output compressed h5 is to be kept in the repo.

"""
import pandas as pd
from mth5_test_data.util import MTH5_TEST_DATA_DIR as DATA_DIR
from aurora.general_helper_functions import TEST_PATH

RAW_DATA_DIR = DATA_DIR.joinpath("iris/BK/2004/ATS")


TARGET_DATA_DIR = TEST_PATH.joinpath("parkfield", "data")
merged_h5 = TARGET_DATA_DIR.joinpath("pkd_sao_2004_272_00-02.h5")

channel_keys = ["hx", "hy", "ex", "ey"]
pkd_df = pd.read_csv(RAW_DATA_DIR.joinpath("PKD_272_00.csv"))
pkd_df = pkd_df[channel_keys]
sao_df = pd.read_csv(RAW_DATA_DIR.joinpath("SAO_272_00.csv"))
sao_df = sao_df[channel_keys]

h5_fn = merged_h5  # "test.h5"


for ch in channel_keys:
    pkd_df[ch].to_hdf(h5_fn, f"{ch}_pkd", complib="zlib", complevel=5)
    sao_df[ch].to_hdf(h5_fn, f"{ch}_sao", complib="zlib", complevel=5)
pkd_df_hx = pd.read_hdf(h5_fn, "hx_pkd")

with pd.HDFStore(merged_h5) as hdf:
    # This prints a list of all group names:
    print(hdf.keys())



from aurora.config import BANDS_DEFAULT_FILE
from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.test_utils.parkfield.make_parkfield_mth5 import ensure_h5_exists
from aurora.test_utils.parkfield.path_helpers import AURORA_RESULTS_PATH

# from aurora.test_utils.parkfield.path_helpers import CONFIG_PATH
from aurora.test_utils.parkfield.path_helpers import DATA_PATH
from aurora.test_utils.parkfield.path_helpers import EMTF_RESULTS_PATH
from aurora.transfer_function.kernel_dataset import KernelDataset
from aurora.transfer_function.plot.comparison_plots import compare_two_z_files

from mth5.helpers import close_open_files


def test_processing(z_file_path=None, test_clock_zero=False):
    """
    Parameters
    ----------
    z_file_path: str or Path or None
        Where to store zfile

    Returns
    -------
    tf_cls: mt_metadata.transfer_functions.core.TF
        The TF object,

    """
    close_open_files()
    h5_path = ensure_h5_exists()

    run_summary = RunSummary()
    h5s_list = [
        h5_path,
    ]
    run_summary.from_mth5s(h5s_list)
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary, "PKD")

    cc = ConfigCreator()
    config = cc.create_from_kernel_dataset(
        tfk_dataset,
        estimator={"engine": "RME"},
        output_channels=["ex", "ey"],
    )

    if test_clock_zero:
        for dec_lvl_cfg in config.decimations:
            dec_lvl_cfg.window.clock_zero_type = test_clock_zero
            if test_clock_zero == "user specified":
                dec_lvl_cfg.window.clock_zero = "2004-09-28 00:00:10+00:00"

    show_plot = False
    tf_cls = process_mth5(
        config,
        tfk_dataset,
        units="MT",
        show_plot=show_plot,
        z_file_path=z_file_path,
    )

    tf_cls.write(fn="emtfxml_test.xml", file_type="emtfxml")
    return tf_cls


def test():
    import logging

    logging.getLogger("matplotlib.font_manager").disabled = True
    logging.getLogger("matplotlib.ticker").disabled = True

    z_file_path = AURORA_RESULTS_PATH.joinpath("pkd.zss")
    tf_cls = test_processing(z_file_path=z_file_path)
    tf_cls.write("pkd_mt_metadata.zss", file_type="zss")
    test_processing(
        z_file_path=z_file_path,
        test_clock_zero="user specified",
    )
    test_processing(z_file_path=z_file_path, test_clock_zero="data start")

    # COMPARE WITH ARCHIVED Z-FILE
    auxilliary_z_file = EMTF_RESULTS_PATH.joinpath("PKD_272_00.zrr")
    if z_file_path.exists():
        compare_two_z_files(
            z_file_path,
            auxilliary_z_file,
            label1="aurora",
            label2="emtf",
            scale_factor1=1,
            out_file="SS.png",
            markersize=3,
            rho_ylims=[1e0, 1e3],
            xlims=[0.05, 500],
        )
    else:
        print("Z-File not found - Parkfield tests failed to generate output")
        print("NCEDC probably not returning data")


if __name__ == "__main__":
    test()



#import logging
import unittest

from aurora.pipelines.run_summary import RunSummary
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test12rr_h5
from aurora.test_utils.synthetic.paths import DATA_PATH


class TestRunSummary(unittest.TestCase):
    """ """
    @classmethod
    def setUpClass(self):
        self._mth5_path = DATA_PATH.joinpath("test12rr.h5")
        if not self._mth5_path.exists():
            self._mth5_path = create_test12rr_h5()
        self._rs = RunSummary()
        self._rs.from_mth5s([self._mth5_path,])
        print("OK")

    def setUp(self):
        rs = self._rs.clone()

    def test_add_duration(self):
        rs = self._rs.clone()
        rs.add_duration()
        assert("duration" in rs.df.columns)




def main():
    # tmp = TestRunSummary()
    # tmp.setUpClass()
    # tmp.test_add_duration()
    unittest.main()


if __name__ == "__main__":
    main()



from deprecated import deprecated
import numpy as np
import pandas as pd
import scipy.signal as ssig
import xarray as xr

from aurora.time_series.windowed_time_series import WindowedTimeSeries
from aurora.time_series.windowing_scheme import window_scheme_from_decimation

def validate_sample_rate(run_ts, expected_sample_rate, tol=1e-4):
    """

    Parameters
    ----------
    run_ts: mth5.timeseries.run_ts.RunTS
        Time series object
    expected_sample_rate: float
        The samepling rate the time series is expected to have. Normally taken from
        the processing config

    """
    if run_ts.sample_rate != expected_sample_rate:
        print(
            f"sample rate in run time series {run_ts.sample_rate} and "
            f"processing decimation_obj {expected_sample_rate} do not match"
        )
        delta = run_ts.sample_rate - expected_sample_rate
        if np.abs(delta) > tol:
            print(f"Delta sample rate {delta} > {tol} tolerance")
            print("TOL should be a percentage")
            raise Exception


def apply_prewhitening(decimation_obj, run_xrds_input):
    """
    Applys prewhitening to time series to avoid spectral leakage when FFT is applied.

    If "first difference", may want to consider clipping first and last sample from
    the differentiated time series.

    Parameters
    ----------
    decimation_obj : mt_metadata.transfer_functions.processing.aurora.DecimationLevel
        Information about how the decimation level is to be processed
    run_xrds_input : xarray.core.dataset.Dataset
        Time series to be prewhitened

    Returns
    -------
    run_xrds : xarray.core.dataset.Dataset
        prewhitened time series

    """
    if not decimation_obj.prewhitening_type:
        return run_xrds_input

    if decimation_obj.prewhitening_type == "first difference":
        run_xrds = run_xrds_input.differentiate("time")

    else:
        print(f"{decimation_obj.prewhitening_type} pre-whitening not implemented")
        raise NotImplementedError
    return run_xrds


def apply_recoloring(decimation_obj, stft_obj):
    """
    Parameters
    ----------
    decimation_obj : mt_metadata.transfer_functions.processing.fourier_coefficients.decimation.Decimation
        Information about how the decimation level is to be processed
    stft_obj : xarray.core.dataset.Dataset
        Time series of Fourier coefficients to be recoloured


    Returns
    -------
    stft_obj : xarray.core.dataset.Dataset
        Recolored time series of Fourier coefficients
    """
    # No recoloring needed if prewhitening not appiled, or recoloring set to False
    if not decimation_obj.prewhitening_type:
        return stft_obj
    if not decimation_obj.recoloring:
        return stft_obj

    if decimation_obj.prewhitening_type == "first difference":
        freqs = decimation_obj.fft_frequencies
        prewhitening_correction = 1.0j * 2 * np.pi * freqs  # jw

        stft_obj /= prewhitening_correction

        # suppress nan and inf to mute later warnings
        if prewhitening_correction[0] == 0.0:
            cond = stft_obj.frequency != 0.0
            stft_obj = stft_obj.where(cond, complex(0.0))
    # elif decimation_obj.prewhitening_type == "ARMA":
    #     from statsmodels.tsa.arima.model import ARIMA
    #     AR = 3 # add this to processing config
    #     MA = 4 # add this to processing config

    else:
        print(f"{decimation_obj.prewhitening_type} recoloring not yet implemented")
        raise NotImplementedError

    return stft_obj


def run_ts_to_stft_scipy(decimation_obj, run_xrds_orig):
    """
    Parameters
    ----------
    decimation_obj : mt_metadata.transfer_functions.processing.aurora.DecimationLevel
        Information about how the decimation level is to be processed
    run_xrds_orig : : xarray.core.dataset.Dataset
        Time series to be processed

    Returns
    -------
    stft_obj : xarray.core.dataset.Dataset
        Time series of Fourier coefficients
    """
    run_xrds = apply_prewhitening(decimation_obj, run_xrds_orig)
    windowing_scheme = window_scheme_from_decimation(decimation_obj)

    stft_obj = xr.Dataset()
    for channel_id in run_xrds.data_vars:
        ff, tt, specgm = ssig.spectrogram(
            run_xrds[channel_id].data,
            fs=decimation_obj.sample_rate_decimation,
            window=windowing_scheme.taper,
            nperseg=decimation_obj.window.num_samples,
            noverlap=decimation_obj.window.overlap,
            detrend="linear",
            scaling="density",
            mode="complex",
        )

        # drop Nyquist>
        ff = ff[:-1]
        specgm = specgm[:-1, :]
        specgm *= np.sqrt(2)

        # make time_axis
        tt = tt - tt[0]
        tt *= decimation_obj.sample_rate_decimation
        time_axis = run_xrds.time.data[tt.astype(int)]

        xrd = xr.DataArray(
            specgm.T,
            dims=["time", "frequency"],
            coords={"frequency": ff, "time": time_axis},
        )
        stft_obj.update({channel_id: xrd})

    stft_obj = apply_recoloring(decimation_obj, stft_obj)

    return stft_obj


def truncate_to_clock_zero(decimation_obj, run_xrds):
    """
    Compute the time interval between the first data sample and the clockzero
    Identify the first sample in the xarray time series that corresponds to a
    window start sample.

    Parameters
    ----------
    decimation_obj: mt_metadata.transfer_functions.processing.aurora.DecimationLevel
        Information about how the decimation level is to be processed
    run_xrds : xarray.core.dataset.Dataset
        normally extracted from mth5.RunTS


    Returns
    -------
    run_xrds : xarray.core.dataset.Dataset
        same as the input time series, but possibly slightly shortened
    """
    if decimation_obj.window.clock_zero_type == "ignore":
        pass
    else:
        clock_zero = pd.Timestamp(decimation_obj.window.clock_zero)
        clock_zero = clock_zero.to_datetime64()
        delta_t = clock_zero - run_xrds.time[0]
        assert delta_t.dtype == "<m8[ns]"  # expected in nanoseconds
        delta_t_seconds = int(delta_t) / 1e9
        if delta_t_seconds == 0:
            pass  # time series start is already clock zero
        else:
            windowing_scheme = window_scheme_from_decimation(decimation_obj)
            number_of_steps = delta_t_seconds / windowing_scheme.duration_advance
            n_partial_steps = number_of_steps - np.floor(number_of_steps)
            n_clip = n_partial_steps * windowing_scheme.num_samples_advance
            n_clip = int(np.round(n_clip))
            t_clip = run_xrds.time[n_clip]
            cond1 = run_xrds.time >= t_clip
            print(
                f"dropping {n_clip} samples to agree with "
                f"{decimation_obj.window.clock_zero_type} clock zero {clock_zero}"
            )
            run_xrds = run_xrds.where(cond1, drop=True)
    return run_xrds


def nan_to_mean(xrds):
    """
    Set Nan values to mean value

    :param xrds: DESCRIPTION
    :type xrds: TYPE
    :return: DESCRIPTION
    :rtype: TYPE

    """
    for ch in xrds.keys():
        null_values_present = xrds[ch].isnull().any()
        if null_values_present:
            print(
                "Null values detected in xrds -- this is not expected and should be examined"
            )
            value = np.nan_to_num(np.nanmean(xrds[ch].data))
            xrds[ch] = xrds[ch].fillna(value)
    return xrds


def run_ts_to_stft(decimation_obj, run_xrds_orig):
    """

    Parameters
    ----------
    decimation_obj : mt_metadata.transfer_functions.processing.aurora.DecimationLevel
        Information about how the decimation level is to be processed
    run_ts : xarray.core.dataset.Dataset
        normally extracted from mth5.RunTS

    Returns
    -------
    stft_obj: xarray.core.dataset.Dataset
        Note that the STFT object may have inf/nan in the DC harmonic, introduced by
        recoloring. This really doesn't matter since we don't use the DC harmonic for
        anything.
    """
    # need to remove any nans before windowing, or else if there is a single
    # nan then the whole channel becomes nan.
    run_xrds = nan_to_mean(run_xrds_orig)
    run_xrds = apply_prewhitening(decimation_obj, run_xrds_orig)
    run_xrds = truncate_to_clock_zero(decimation_obj, run_xrds)
    windowing_scheme = window_scheme_from_decimation(decimation_obj)
    windowed_obj = windowing_scheme.apply_sliding_window(
        run_xrds, dt=1.0 / decimation_obj.sample_rate_decimation
    )
    if not np.prod(windowed_obj.to_array().data.shape):
        raise ValueError

    windowed_obj = WindowedTimeSeries.detrend(data=windowed_obj, detrend_type="linear")
    tapered_obj = windowed_obj * windowing_scheme.taper
    stft_obj = windowing_scheme.apply_fft(
        tapered_obj, detrend_type=decimation_obj.extra_pre_fft_detrend_type
    )
    stft_obj = apply_recoloring(decimation_obj, stft_obj)

    return stft_obj


def calibrate_stft_obj(stft_obj, run_obj, units="MT", channel_scale_factors=None):
    """

    Parameters
    ----------
    stft_obj : xarray.core.dataset.Dataset
        Time series of Fourier coefficients to be calibrated
    run_obj : mth5.groups.master_station_run_channel.RunGroup
        Provides information about filters for calibration
    units : string
        usually "MT", contemplating supporting "SI"
    scale_factors : dict or None
        keyed by channel, supports a single scalar to apply to that channels data
        Useful for debugging.  Should not be used in production and should throw a
        warning if it is not None

    Returns
    -------
    stft_obj : xarray.core.dataset.Dataset
        Time series of calibrated Fourier coefficients
    """
    for channel_id in stft_obj.keys():
        mth5_channel = run_obj.get_channel(channel_id)
        channel_filter = mth5_channel.channel_response_filter
        if not channel_filter.filters_list:
            print("WARNING UNEXPECTED CHANNEL WITH NO FILTERS")
            if channel_id == "hy":
                print("Channel HY has no filters, try using filters from HX")
                channel_filter = run_obj.get_channel("hx").channel_response_filter
        calibration_response = channel_filter.complex_response(stft_obj.frequency.data)
        if channel_scale_factors:
            try:
                channel_scale_factor = channel_scale_factors[channel_id]
            except KeyError:
                channel_scale_factor = 1.0
            calibration_response /= channel_scale_factor
        if units == "SI":
            print("Warning: SI Units are not robustly supported issue #36")

        stft_obj[channel_id].data /= calibration_response
    return stft_obj


def prototype_decimate(config, run_xrds):
    """
    Consider:
    1. Moving this function into time_series/decimate.py
    2. Replacing the downsampled_time_axis with rolling mean, or somthing that takes
    the average value of the time, not the window start

    Parameters
    ----------
    config : mt_metadata.transfer_functions.processing.aurora.Decimation
    run_xrds: xr.Dataset
        Originally from mth5.timeseries.run_ts.RunTS.dataset, but possibly decimated
        multiple times

    Returns
    -------
    xr_ds: xr.Dataset
        Decimated version of the input run_xrds
    """
    slicer = slice(None, None, int(config.factor))  # decimation.factor
    downsampled_time_axis = run_xrds.time.data[slicer]

    num_observations = len(downsampled_time_axis)
    channel_labels = list(run_xrds.data_vars.keys())  # run_ts.channels
    num_channels = len(channel_labels)
    new_data = np.full((num_observations, num_channels), np.nan)
    for i_ch, ch_label in enumerate(channel_labels):
        new_data[:, i_ch] = ssig.decimate(run_xrds[ch_label], int(config.factor))

    xr_da = xr.DataArray(
        new_data,
        dims=["time", "channel"],
        coords={"time": downsampled_time_axis, "channel": channel_labels},
    )
    attr_dict = run_xrds.attrs
    attr_dict["sample_rate"] = config.sample_rate
    xr_da.attrs = attr_dict
    xr_ds = xr_da.to_dataset("channel")
    return xr_ds


def prototype_decimate_2(config, run_xrds):
    """
    Uses the built-in xarray coarsen method.   Not clear what AAF effects are.
    Method is fast.  Might be non-linear.  Seems to give similar performance to
    prototype_decimate for synthetic data.

    N.B. config.factor must be integer valued

    Parameters
    ----------
    config : mt_metadata.transfer_functions.processing.aurora.Decimation
    run_xrds: xr.Dataset
        Originally from mth5.timeseries.run_ts.RunTS.dataset, but possibly decimated
        multiple times

    Returns
    -------
    xr_ds: xr.Dataset
        Decimated version of the input run_xrds
    """
    new_xr_ds = run_xrds.coarsen(time=int(config.factor), boundary="trim").mean()
    attr_dict = run_xrds.attrs
    attr_dict["sample_rate"] = config.sample_rate
    new_xr_ds.attrs = attr_dict
    return new_xr_ds


def prototype_decimate_3(config, run_xrds):
    """
    Uses the built-in xarray coarsen method.   Not clear what AAF effects are.
    Method is fast.  Might be non-linear.  Seems to give similar performance to
    prototype_decimate for synthetic data.

    N.B. config.factor must be integer valued

    Parameters
    ----------
    config : mt_metadata.transfer_functions.processing.aurora.Decimation
    run_xrds: xr.Dataset
        Originally from mth5.timeseries.run_ts.RunTS.dataset, but possibly decimated
        multiple times

    Returns
    -------
    xr_ds: xr.Dataset
        Decimated version of the input run_xrds
    """
    dt = run_xrds.time.diff(dim="time").median().values
    dt_new = config.factor * dt
    dt_new = dt_new.__str__().replace("nanoseconds", "ns")
    new_xr_ds = run_xrds.resample(time=dt_new).mean(dim="time")
    attr_dict = run_xrds.attrs
    attr_dict["sample_rate"] = config.sample_rate
    new_xr_ds.attrs = attr_dict
    return new_xr_ds



"""
Notes in google doc:
https://docs.google.com/document/d/1CsRhSLXsRG8HQxM4lKNqVj-V9KA9iUQAvCOtouVzFs0/edit?usp=sharing


"""
import numpy as np
from numpy.lib.stride_tricks import as_strided
from numba import jit
import time


# Window-to-timeseries relationshp
def available_number_of_windows_in_array(n_samples_array, n_samples_window, n_advance):
    """

    Parameters
    ----------
    n_samples_array: int
        The length of the time series
    n_samples_window: int
        The length of the window (in samples)
    n_advance: int
        The number of samples the window advances at each step

    Returns
    -------
    available_number_of_strides: int
        The number of windows the time series will yield
    """
    stridable_samples = n_samples_array - n_samples_window
    if stridable_samples < 0:
        print("CRITICAL Window is longer than the time series")
        return 0
    available_number_of_strides = int(np.floor(stridable_samples / n_advance))
    available_number_of_strides += 1
    return available_number_of_strides


# Sliding Window Operators
def sliding_window_crude(
    data, num_samples_window, num_samples_advance, num_windows=None
):
    """

    Parameters
    ----------
    data: np.ndarray
        The time series data to be windowed
    num_samples_window: int
        The length of the window (in samples)
    num_samples_advance: int
        The number of samples the window advances at each step
    num_windows: int
        The number of windows to "take".  Must be less or equal to the number of
        available windows.

    Returns
    -------
    output_array: numpy.ndarray
        The windowed time series
    """
    if num_windows is None:
        num_windows = available_number_of_windows_in_array(
            len(data), num_samples_window, num_samples_advance
        )
    output_array = np.full((num_windows, num_samples_window), np.nan)
    for i in range(num_windows):
        output_array[i, :] = data[
            i * num_samples_advance : i * num_samples_advance + num_samples_window
        ]

    return output_array


@jit
def sliding_window_numba(data, num_samples_window, num_samples_advance, num_windows):
    """

    Parameters
    ----------
    data: np.ndarray
        The time series data to be windowed
    num_samples_window: int
        The length of the window (in samples)
    num_samples_advance: int
        The number of samples the window advances at each step
    num_windows: int
        The number of windows to "take".

    Returns
    -------
    output_array: numpy.ndarray
        The windowed time series
    """
    output_array = np.full((num_windows, num_samples_window), np.nan)
    for i in range(num_windows):
        output_array[i, :] = data[
            i * num_samples_advance : i * num_samples_advance + num_samples_window
        ]

    return output_array


def striding_window(data, num_samples_window, num_samples_advance, num_windows=None):
    """
    Applies a striding window to an array.  We use 1D arrays here.
    Note that this method is extendable to N-dimensional arrays as was once shown
    at  http://www.johnvinyard.com/blog/?p=268

    Karl has an implementation of this code but chose to restict to 1D here.
    This is becuase of several warnings encountered, on the notes of stride_tricks.py,
    as well as for example here:
    https://stackoverflow.com/questions/4936620/using-strides-for-an-efficient-moving-average-filter

    While we can possibly setup Aurora so that no copies of the strided window are made
    downstream, we cannot guarantee that another user may not add methods that require
    copies.  For robustness we will use 1d implementation only for now.

    Another clean example of this method can be found in the razorback codes from brgm.

    result is 2d: result[i] is the i-th window

    >>> sliding_window(np.arange(15), 4, 3, 2)
    array([[0, 1, 2],
           [2, 3, 4],
           [4, 5, 6],
           [6, 7, 8]])

    Parameters
    ----------
    data: np.ndarray
        The time series data to be windowed
    num_samples_window: int
        The length of the window (in samples)
    num_samples_advance: int
        The number of samples the window advances at each step
    num_windows: int
        The number of windows to "take".  Must be less or equal to the number of
        available windows.

    Returns
    -------
    strided_window: numpy.ndarray
        The windowed time series

    """
    if num_windows is None:
        num_windows = available_number_of_windows_in_array(
            len(data), num_samples_window, num_samples_advance
        )
    # min_ = (num_windows - 1) * num_samples_advance + num_samples_window
    bytes_per_element = data.itemsize
    output_shape = (num_windows, num_samples_window)
    # print("output_shape", output_shape)
    strides_shape = (num_samples_advance * bytes_per_element, bytes_per_element)
    # strides_shape = None
    print("strides_shape", strides_shape)
    strided_window = as_strided(
        data, shape=output_shape, strides=strides_shape
    )  # , writeable=False)
    return strided_window


# Sliding Window Functions
SLIDING_WINDOW_FUNCTIONS = {
    "crude": sliding_window_crude,
    "numba": sliding_window_numba,
    "stride": striding_window,
}


# FFT Helpers
def apply_fft_to_windowed_array(windowed_array):
    """
    This will operate row-wise as well
    Parameters
    ----------
    windowed_array

    Returns
    -------

    """
    pass


def check_all_sliding_window_functions_are_equivalent():
    """
    simple sanity check that runs each sliding window function on a small array and
    confirms the results are numerically identical.
    Note that striding window will return int types where others return float.
    Returns
    -------

    """

    N = 15
    L = 3
    V = 1
    A = L - V
    data = np.arange(N)
    n_win = available_number_of_windows_in_array(N, L, A)
    results = {}
    for function_label, function in SLIDING_WINDOW_FUNCTIONS.items():
        results[function_label] = function(data, L, A, n_win)
        print(results[function_label])

    for i, function_label in enumerate(results.keys()):
        if i == 0:
            reference_result = results[function_label]
        else:
            difference = reference_result - results[function_label]
            if np.sum(np.abs(difference)) == 0:
                print(f"OK {i}")
                # put an assert here instead of OK


def do_some_tests():
    N = 10000000
    n_samples_window = 128
    n_overlap = 96
    n_advance = n_samples_window - n_overlap

    sw = striding_window(np.arange(15), 3, 2, num_windows=4)
    print(sw)

    t0 = time.time()
    strided_window = striding_window(
        1.0 * np.arange(N), n_samples_window, n_advance
    )  # , num_windows=4)
    strided_window += 1
    print("stride {}".format(time.time() - t0))

    print(strided_window)

    t0 = time.time()
    slid_window = sliding_window_crude(
        1.0 * np.arange(N), n_samples_window, n_advance
    )  # , num_windows=4)
    slid_window += 1
    print("crude  {}".format(time.time() - t0))

    print(slid_window)

    num_windows = available_number_of_windows_in_array(N, n_samples_window, n_advance)
    print(num_windows)
    t0 = time.time()
    numba_slid_window = sliding_window_numba(
        1.0 * np.arange(N), n_samples_window, n_advance, num_windows
    )  # , num_windows=4)
    numba_slid_window += 1
    print("numba  {}".format(time.time() - t0))

    t0 = time.time()
    numba_slid_window = sliding_window_numba(
        1.0 * np.arange(N), n_samples_window, n_advance, num_windows
    )  # , num_windows=4)
    print("numba  {}".format(time.time() - t0))


def test_apply_taper():
    import matplotlib.pyplot as plt
    import scipy.signal as ssig

    num_samples_window = 64
    num_windows = 100
    windowed_data = np.abs(np.random.randn(num_windows, num_samples_window))
    taper = ssig.windows.hann(num_samples_window)
    tapered_windowed_data = windowed_data * taper
    plt.plot(windowed_data[0], "r", label="data")
    plt.plot(tapered_windowed_data[0], "g", label="tapered data")
    plt.legend()
    plt.show()
    return


def main():
    check_all_sliding_window_functions_are_equivalent()
    do_some_tests()
    test_apply_taper()


if __name__ == "__main__":
    main()


# """
#     This function was originally copied from:
#         http://www.johnvinyard.com/blog/?p=268
# """
#
# def norm_shape(shape):
#     '''
#     Normalize numpy array shapes so they're always expressed as a tuple,
#     even for one-dimensional shapes.
#
#     Parameters
#         shape - an int, or a tuple of ints
#
#     Returns
#         a shape tuple
#     @note: 20140221, kkappler: this method taken from
#     http://www.johnvinyard.com/blog/?p=268, and used by ensemblizedTimeSeries
#     but probably will be used by other methods.
#     It is less a signal processing method and more a numpy-array-handling
#     tool, in case we ever get that specific with files/functions.
#
#     '''
#     try:
#         i = int(shape)
#         return (i, )
#     except TypeError:
#         # shape was not a number
#         pass
#
#     try:
#         t = tuple(shape)
#         return t
#     except TypeError:
#         # shape was not iterable
#         pass
#
#     raise TypeError('shape must be an int, or a tuple of ints')
#
#
# #</Adding these methods which were developed at QF and needed for first version ops>
# def sliding_window(a, ws, ss=None, flatten=True):
#     '''
#     This function was originally copied from:
#         http://www.johnvinyard.com/blog/?p=268
#
#     usage: sw = sliding_window(a,ws,ss = None,flatten = True):
#     Return a sliding window over a in any number of dimensions
#
#     Parameters:
#         a  - an n-dimensional numpy array
#         ws - an int (a is 1D) or tuple (a is 2D or greater) representing the size
#              of each dimension of the window
#         ss - an int (a is 1D) or tuple (a is 2D or greater) representing the
#              amount to slide the window in each dimension. If not specified, it
#              defaults to ws.
#         flatten - if True, all slices are flattened, otherwise, there is an
#                   extra dimension for each dimension of the input.
#
#     Returns
#         Array, if 2d dimensions are (num_windows, n_points_per_window)
#
#         An array containing each n-dimensional window from a
#         The Ensemblized time series seems in many ways like a MVTS, but it is NOT.
#     It looks like one, because the data are a 2D array, with time moving along the
#     "row" or zero axis, but the similarities stop there.
#     The time axis is different for each row!  Each row spans the same duration,
#     but the actual interval t0+row_duration is difference for each row,, i.e. the
#     t0 is different for each row.
#
#     In the case where the window length is 1, we obntain the special case where
#     the ensemblized time series is just the transpose (not-conjugate) of the input
#     time series.
#
#     There is an excellent article on ensemblization in numpy
#     http://www.johnvinyard.com/blog/?p=268
#
#     @note: For many ensemble processing applications you are NOT obliged
#     to hold the ensembles in memory at a given time.  The old
#     ensembleEdges class was designed with this in mind and probably would
#     be slightly less RAM-heavy.  Prior to 2011 I always created ensemblizedTimeSeries
#     in my codes (even if they were not called by that name) because of the
#     inate support for vector math
#     @note: the ws here is the distance the window slides whcih is L-V, i.e. it
#     is not the Overlap, but rather its difference from L.
#     '''
#     if None is ss:
#         # ss was not provided. the windows will not overlap in any direction.
#         ss = ws
#     ws = norm_shape(ws)
#     ss = norm_shape(ss)
#
#     # convert ws, ss, and a.shape to numpy arrays so that we can do math in every
#     # dimension at once.
#     ws = np.array(ws)
#     ss = np.array(ss)
#     shape = np.array(a.shape)
#
#     # ensure that ws, ss, and a.shape all have the same number of dimensions
#     ls = [len(shape), len(ws), len(ss)]
#     if len(set(ls)) != 1:
#         raise ValueError(\
#         'a.shape, ws and ss must all have the same length. They were %s' % str(ls))
#
#     # ensure that ws is smaller than a in every dimension
#     if np.any(ws > shape):
#         raise ValueError(\
#         'ws cannot be larger than a in any dimension.\
#  a.shape was %s and ws was %s' % (str(a.shape), str(ws)))
#
#     # how many slices will there be in each dimension?
#     newshape = norm_shape(((shape - ws) // ss) + 1)
#     # the shape of the strided array will be the number of slices in each dimension
#     # plus the shape of the window (tuple addition)
#     newshape += norm_shape(ws)
#     # the strides tuple will be the array's strides multiplied by step size, plus
#     # the array's strides (tuple addition)
#     newstrides = norm_shape(np.array(a.strides) * ss) + a.strides
#     strided = ast(a, shape=newshape, strides=newstrides)
#     if not flatten:
#         return strided
#
#     # Collapse strided so that it has one more dimension than the window.  I.e.,
#     # the new array is a flat list of slices.
#     meat = len(ws) if ws.shape else 0
#     firstdim = (np.product(newshape[:-meat]),) if ws.shape else ()
#     dim = np.array(firstdim + (newshape[-meat:]))
#     # remove any dimensions with size 1
#     dim = dim[dim != 1]
#     return strided.reshape(dim)



import numpy as np
import xarray as xr
import unittest

from aurora.time_series.time_axis_helpers import make_time_axis
from aurora.time_series.windowing_scheme import WindowingScheme

np.random.seed(0)


# =============================================================================
#  Helper functions
# =============================================================================


def get_windowing_scheme(
    num_samples_window=32,
    num_samples_overlap=8,
    sample_rate=None,
    taper_family="hamming",
):
    windowing_scheme = WindowingScheme(
        num_samples_window=num_samples_window,
        num_samples_overlap=num_samples_overlap,
        taper_family=taper_family,
        sample_rate=sample_rate,
    )
    return windowing_scheme


def get_xarray_dataset(N=1000, sps=50.0):
    """
    make a few xarrays, then bind them into a dataset
    ToDo: Consider moving this method into test_utils/
    ToDo: Add a test that confirms we get an AttributeError when we try
    to overwrite attrs of xarray (
    try:
        ds.sample_rate=10
        print("was not expecting to be able to overwrite attr of xarray")
        assert(False)
    except AttributeError:
        assert(True)
    """
    t0 = np.datetime64("1977-03-02 12:34:56")
    time_vector = make_time_axis(t0, N, sps)
    ds = xr.Dataset(
        {
            "hx": (
                [
                    "time",
                ],
                np.abs(np.random.randn(N)),
            ),
            "hy": (
                [
                    "time",
                ],
                np.abs(np.random.randn(N)),
            ),
        },
        coords={
            "time": time_vector,
        },
        attrs={
            "some random info": "dogs",
            "some more random info": "cats",
            "sample_rate": sps,
        },
    )
    return ds


# =============================================================================
#  Tests
# =============================================================================


class TestWindowingScheme(unittest.TestCase):
    def setUp(self):
        self.defaut_num_samples_data = 10000
        self.defaut_num_samples_window = 64
        self.default_num_samples_overlap = 50

    def test_cant_write_xarray_attrs(self):
        """
        This could go into a separate module for testing xarray stuff
        """
        ds = get_xarray_dataset()
        try:
            ds.sample_rate = 10
            print("was not expecting to be able to overwrite attr of xarray")
            assert False
        except AttributeError:
            assert True

    def test_instantiate_windowing_scheme(self):
        num_samples_window = 128
        num_samples_overlap = 32
        num_samples_data = 1000
        sample_rate = 50.0
        taper_family = "hamming"
        ws = WindowingScheme(
            num_samples_window=num_samples_window,
            num_samples_overlap=num_samples_overlap,
            num_samples_data=num_samples_data,
            taper_family=taper_family,
        )
        ws.sample_rate = sample_rate
        expected_window_duration = num_samples_window / sample_rate
        assert ws.window_duration == expected_window_duration

    def test_apply_sliding_window(self):
        num_samples_data = self.defaut_num_samples_data
        num_samples_window = self.defaut_num_samples_window
        num_samples_overlap = self.default_num_samples_overlap
        ts = np.random.random(num_samples_data)
        windowing_scheme = WindowingScheme(
            num_samples_window=num_samples_window,
            num_samples_overlap=num_samples_overlap,
        )
        windowed_array = windowing_scheme.apply_sliding_window(ts)
        return windowed_array

    def test_apply_sliding_window_can_return_xarray(self):
        ts = np.arange(15)
        windowing_scheme = WindowingScheme(num_samples_window=3, num_samples_overlap=1)
        windowed_xr = windowing_scheme.apply_sliding_window(ts, return_xarray=True)
        assert isinstance(windowed_xr, xr.DataArray)
        return windowed_xr

    def test_apply_sliding_window_to_xarray(self, return_xarray=False):
        num_samples_data = self.defaut_num_samples_data
        num_samples_window = self.defaut_num_samples_window
        num_samples_overlap = self.default_num_samples_overlap
        xrd = xr.DataArray(
            np.random.randn(num_samples_data, 1),
            dims=["time", "channel"],
            coords={"time": np.arange(num_samples_data)},
        )
        windowing_scheme = WindowingScheme(
            num_samples_window=num_samples_window,
            num_samples_overlap=num_samples_overlap,
        )
        windowed_xrda = windowing_scheme.apply_sliding_window(
            xrd, return_xarray=return_xarray
        )
        return windowed_xrda

    def test_can_apply_taper(self):
        from aurora.time_series.window_helpers import (
            available_number_of_windows_in_array,
        )

        num_samples_data = self.defaut_num_samples_data
        num_samples_window = self.defaut_num_samples_window
        num_samples_overlap = self.default_num_samples_overlap
        ts = np.random.random(num_samples_data)
        windowing_scheme = WindowingScheme(
            num_samples_window=num_samples_window,
            num_samples_overlap=num_samples_overlap,
            taper_family="hamming",
        )
        expected_advance = num_samples_window - num_samples_overlap
        assert windowing_scheme.num_samples_advance == expected_advance
        expected_num_windows = available_number_of_windows_in_array(
            num_samples_data, num_samples_window, expected_advance
        )
        num_windows = windowing_scheme.available_number_of_windows(num_samples_data)
        assert num_windows == expected_num_windows
        windowed_data = windowing_scheme.apply_sliding_window(ts)
        tapered_windowed_data = windowing_scheme.apply_taper(windowed_data)
        assert (windowed_data[:, 0] != tapered_windowed_data[:, 0]).all()

        # import matplotlib.pyplot as plt
        # plt.plot(windowed_data[0],'r');plt.plot(tapered_windowed_data[0],'g')
        # plt.show()
        return

    def test_taper_dataset(self, plot=False):
        import matplotlib.pyplot as plt

        windowing_scheme = get_windowing_scheme(
            num_samples_window=64,
            num_samples_overlap=8,
            sample_rate=None,
            taper_family="hamming",
        )
        ds = get_xarray_dataset()

        windowed_dataset = windowing_scheme.apply_sliding_window(ds, return_xarray=True)
        if plot:
            fig, ax = plt.subplots()
            ax.plot(windowed_dataset["hx"].data[0, :], "r", label="window0")
            ax.plot(windowed_dataset["hx"].data[1, :], "r", label="window1")
        tapered_dataset = windowing_scheme.apply_taper(windowed_dataset)
        if plot:
            ax.plot(tapered_dataset["hx"].data[0, :], "g", label="tapered0")
            ax.plot(tapered_dataset["hx"].data[1, :], "g", label="tapered1")
            ax.legend()
            plt.show()

    def test_can_create_xarray_dataset_from_several_sliding_window_xarrays(self):
        """
        This method operates on an xarray dataset.
        Returns
        -------
        """
        windowing_scheme = get_windowing_scheme(
            num_samples_window=32, num_samples_overlap=8
        )
        ds = get_xarray_dataset()
        wds = windowing_scheme.apply_sliding_window(ds, return_xarray=True)
        return wds

    def test_fourier_transform(self):
        """
        This method gets a windowed time series, applies a taper, and fft
        """
        sample_rate = 40.0
        windowing_scheme = get_windowing_scheme(
            num_samples_window=128, num_samples_overlap=96, sample_rate=sample_rate
        )

        # Test with xr.Dataset
        ds = get_xarray_dataset(N=10000, sps=sample_rate)
        windowed_dataset = windowing_scheme.apply_sliding_window(ds)
        tapered_windowed_dataset = windowing_scheme.apply_taper(windowed_dataset)
        stft = windowing_scheme.apply_fft(tapered_windowed_dataset)
        assert isinstance(stft, xr.Dataset)

        # Test with xr.DataArray
        da = ds.to_array("channel")
        windowed_dataset = windowing_scheme.apply_sliding_window(da)
        tapered_windowed_dataset = windowing_scheme.apply_taper(windowed_dataset)
        stft = windowing_scheme.apply_fft(tapered_windowed_dataset)
        assert isinstance(stft, xr.DataArray)

        # import matplotlib.pyplot as plt
        # plt.plot(stft.frequency.data, np.abs(stft["hx"].data.mean(axis=0)))
        # plt.show()


def main():
    """
    Testing the windowing scheme
    """
    unittest.main()


if __name__ == "__main__":
    main()



from mth5.utils.helpers import initialize_mth5
from aurora.test_utils.synthetic.paths import DATA_PATH
from mth5.helpers import close_open_files

import datetime


def fix_time(tstmp):
    year = tstmp.year
    month = tstmp.month
    day = tstmp.day
    hour = tstmp.hour
    minute = tstmp.minute
    second = tstmp.second
    out = datetime.datetime(year, month, day, hour, minute, second)
    return out


def test_can_slice_a_run_ts_using_timestamp():
    close_open_files()
    mth5_path = DATA_PATH.joinpath("test1.h5")
    mth5_obj = initialize_mth5(mth5_path, "r")
    df = mth5_obj.channel_summary.to_dataframe()
    run_001 = mth5_obj.get_run("test1", "001")
    run_ts_01 = run_001.to_runts()
    start = df.iloc[0].start
    end = df.iloc[0].end
    print(" Workaround for mt_metadata issue #86 - remove once that is resolved")
    start = fix_time(start)
    end = fix_time(end)
    run_ts_02 = run_001.to_runts(start=start, end=end)
    run_ts_03 = run_001.to_runts(
        start=start, end=end + datetime.timedelta(microseconds=499999)
    )

    run_ts_04 = run_001.to_runts(
        start=start, end=end + datetime.timedelta(microseconds=500000)
    )
    print(f"run_ts_01 has {len(run_ts_01.dataset.ex.data)} samples")
    print(f"run_ts_02 has {len(run_ts_02.dataset.ex.data)} samples")
    print(f"run_ts_03 has {len(run_ts_03.dataset.ex.data)} samples")
    print(f"run_ts_04 has {len(run_ts_04.dataset.ex.data)} samples")


def main():
    test_can_slice_a_run_ts_using_timestamp()


if __name__ == "__main__":
    main()



import logging
import unittest
from aurora.config import BANDS_DEFAULT_FILE
from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.run_summary import RunSummary
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test3_h5
from aurora.test_utils.synthetic.paths import AURORA_RESULTS_PATH
from aurora.transfer_function.kernel_dataset import KernelDataset
from mth5.helpers import close_open_files


class TestMultiRunProcessing(unittest.TestCase):
    """
    Runs several synthetic multi-run processing tests from config creation to
    tf_collection.

    """

    remake_mth5_for_each_test = False

    def setUp(self):
        close_open_files()
        logging.getLogger("matplotlib.font_manager").disabled = True
        logging.getLogger("matplotlib.ticker").disabled = True

    @classmethod
    def setUpClass(cls) -> None:
        """ Add a fresh h5 to start the test, sowe don't have FCs in there from other tests"""
        create_test3_h5(force_make_mth5=True)

    def make_mth5(self):
        close_open_files()
        mth5_path = create_test3_h5(force_make_mth5=self.remake_mth5_for_each_test)
        return mth5_path

    def make_run_summary(self):
        mth5_path = self.make_mth5()
        mth5s = [
            mth5_path,
        ]
        run_summary = RunSummary()
        run_summary.from_mth5s(mth5s)
        return run_summary

    def test_each_run_individually(self):
        close_open_files()
        run_summary = self.make_run_summary()
        for run_id in run_summary.df.run_id.unique():
            kernel_dataset = KernelDataset()
            kernel_dataset.from_run_summary(run_summary, "test3")
            station_runs_dict = {}
            station_runs_dict["test3"] = [
                run_id,
            ]
            keep_or_drop = "keep"
            kernel_dataset.select_station_runs(station_runs_dict, keep_or_drop)
            cc = ConfigCreator()
            config = cc.create_from_kernel_dataset(kernel_dataset)

            for decimation in config.decimations:
                decimation.estimator.engine = "RME"
            show_plot = False  # True
            z_file_path = AURORA_RESULTS_PATH.joinpath(f"syn3_{run_id}.zss")
            tf_cls = process_mth5(
                config,
                kernel_dataset,
                units="MT",
                show_plot=show_plot,
                z_file_path=z_file_path,
            )
            xml_file_base = f"syn3_{run_id}.xml"
            xml_file_name = AURORA_RESULTS_PATH.joinpath(xml_file_base)
            tf_cls.write(fn=xml_file_name, file_type="emtfxml")

    def test_all_runs(self):
        close_open_files()
        run_summary = self.make_run_summary()
        kernel_dataset = KernelDataset()
        kernel_dataset.from_run_summary(run_summary, "test3")
        cc = ConfigCreator()
        config = cc.create_from_kernel_dataset(
            kernel_dataset, estimator={"engine": "RME"}
        )

        show_plot = False  # True
        z_file_path = AURORA_RESULTS_PATH.joinpath("syn3_all.zss")
        tf_cls = process_mth5(
            config,
            kernel_dataset,
            units="MT",
            show_plot=show_plot,
            z_file_path=z_file_path,
        )
        xml_file_name = AURORA_RESULTS_PATH.joinpath("syn3_all.xml")
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")

    def test_works_with_truncated_run(self):
        """
        Synthetic runs are 40000s long.  By truncating one of the runs to 10000s,
        we make the 4th decimation invalid for that run invalid.  By truncating to
        2000s long we make the 3rd and 4th decimation levels invalid.
        Returns
        -------

        """
        import pandas as pd

        run_summary = self.make_run_summary()
        delta = pd.Timedelta(seconds=38000)
        run_summary.df.end.iloc[1] -= delta
        kernel_dataset = KernelDataset()
        kernel_dataset.from_run_summary(run_summary, "test3")
        cc = ConfigCreator()
        config = cc.create_from_kernel_dataset(
            kernel_dataset, estimator={"engine": "RME"}
        )

        show_plot = False  # True
        z_file_path = AURORA_RESULTS_PATH.joinpath("syn3_all_truncated_run.zss")
        tf_cls = process_mth5(
            config,
            kernel_dataset,
            units="MT",
            show_plot=show_plot,
            z_file_path=z_file_path,
        )
        xml_file_name = AURORA_RESULTS_PATH.joinpath("syn3_all_truncated_run.xml")
        tf_cls.write(fn=xml_file_name, file_type="emtfxml")


def main():
    # tmp = TestMultiRunProcessing()
    # tmp.setUp()
    # tmp.test_works_with_truncated_run()
    # tmp.test_all_runs()
    unittest.main()


if __name__ == "__main__":
    main()



from matplotlib.gridspec import GridSpec

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as ssig


def is_flat_amplitude(array):
    """
    Check of an amplitude response is basically flat.  If so, it is best to tune the y-axis lims to
    make numeric noise invisible
    Parameters
    ----------
    array
    Returns
    -------
    """
    differences = np.diff(np.abs(array))
    if np.isclose(differences, 0.0).all():
        return True
    else:
        return False


def cast_angular_frequency_to_period_or_hertz(w, units):
    if units.lower() == "period":
        x_axis = (2.0 * np.pi) / w
    elif units.lower() == "frequency":
        x_axis = w / (2.0 * np.pi)
    return x_axis


def plot_complex_response(frequency, complex_response, **kwargs):
    """
    ToDo: add methods for suporting instrument object but for now take as kwargs
    :param frequency: numpy array of frequencies at which complex response is defined
    :param complex_response: numpy array of full complex valued response function
    :param kwargs:
    :return:
    """

    show = kwargs.get("show", True)
    linewidth = kwargs.get("linewidth", 3)

    make = kwargs.get("make", None)
    model = kwargs.get("model", None)
    y_amp_string = kwargs.get("yamp", None)

    amplitude = np.abs(complex_response)
    phase = np.angle(complex_response)

    plt.figure(1)
    # plt.clf()
    ax1 = plt.subplot(2, 1, 1)
    ax1.loglog(frequency, amplitude, linewidth=linewidth)
    ax1.set_title("{}-{}    Amplitude Response".format(make, model))
    ax1.grid(True, which="both", ls="-")
    ax1.set_ylabel("{}".format(y_amp_string))
    y_lim = ax1.get_ylim()
    ax1.set_ylim((y_lim[0], 1.1 * y_lim[1]))
    ax2 = plt.subplot(2, 1, 2)
    ax2.semilogx(frequency, phase, linewidth=linewidth)
    ax2.set_title("{}-{}    Phase Response".format(make, model))
    ax2.grid(True, which="both", ls="-")
    if show:
        plt.show()


def plot_response_pz(
    w_obs=None,
    resp_obs=None,
    zpk_obs=None,
    zpk_pred=None,
    w_values=None,
    xlim=None,
    title=None,
    x_units="Period",
):
    """
    This function was contributed by Ben Murphy at USGS
    2021-03-17: there are some issues encountered when using this function to plot
    generic resposnes, looks like the x-axis gets out of order when using frequency
    as the axis ...

    Parameters
    ----------
    w_obs : numpy array
        Angular frequencies from lab observations, for example a fap calibration table
    resp_obs: numpy array
        Complex valued response associated with lab observation
    zpk_obs: scipy.signal.ltisys.ZerosPolesGainContinuous
        Pole-Zero object representing the lab observation
    zpk_pred: scipy.signal.ltisys.ZerosPolesGainContinuous
        Pole-Zero object representing the model
    w_values: numpy array
        Angular frequencies at which to evaluate zpk_pred and zpk_obs
    xlim:
    title: string
        Title for the plot
    x_units: string
        One of ["Period" , "frequency"], case insensitve

    """
    # set up the plotting axes
    fig = plt.figure(figsize=(14, 4))
    if title is not None:
        fig.suptitle(title)
    gs = GridSpec(2, 3)
    ax_amp = fig.add_subplot(gs[0, :2])
    ax_phs = fig.add_subplot(gs[1, :2])
    ax_pz = fig.add_subplot(gs[:, 2], aspect="equal")

    # plot observed (lab) response as amplitude and phase
    if w_obs is not None and resp_obs is not None:

        response_amplitude = np.absolute(resp_obs)
        if is_flat_amplitude(resp_obs):
            response_amplitude[:] = response_amplitude[0]
            ax_amp.set_ylim([0.9 * response_amplitude[0], 1.1 * response_amplitude[0]])
        x_values = cast_angular_frequency_to_period_or_hertz(w_obs, x_units)
        ax_amp.plot(
            x_values,
            response_amplitude,
            color="tab:blue",
            linewidth=1.5,
            linestyle="-",
            label="True",
        )
        ax_phs.plot(
            x_values,
            np.angle(resp_obs, deg=True),
            color="tab:blue",
            linewidth=1.5,
            linestyle="-",
        )
    elif zpk_obs is not None:
        w_obs, resp_obs = ssig.freqresp(zpk_obs, w=w_values)
        x_values = cast_angular_frequency_to_period_or_hertz(w_obs, x_units)
        ax_amp.plot(
            x_values,
            np.absolute(resp_obs),
            color="tab:blue",
            linewidth=1.5,
            linestyle="-",
            label="True",
        )
        ax_phs.plot(
            x_values,
            np.angle(resp_obs, deg=True),
            color="tab:blue",
            linewidth=1.5,
            linestyle="-",
        )
        ax_pz.scatter(
            np.real(zpk_obs.zeros),
            np.imag(zpk_obs.zeros),
            s=75,
            marker="o",
            ec="tab:blue",
            fc="w",
            label="True Zeros",
        )
        ax_pz.scatter(
            np.real(zpk_obs.poles),
            np.imag(zpk_obs.poles),
            s=75,
            marker="x",
            ec="tab:blue",
            fc="tab:blue",
            label="True Poles",
        )

    # plot predicted response (model) as amplitude and phase
    if zpk_pred is not None:
        w_pred, resp_pred = ssig.freqresp(zpk_pred, w=w_values)
        x_values = cast_angular_frequency_to_period_or_hertz(w_values, x_units)

        ax_amp.plot(
            x_values,
            np.absolute(resp_pred),
            color="tab:red",
            linewidth=3,
            linestyle=":",
            label="Fit",
        )
        # print(np.angle(resp_pred, deg=True))
        ax_phs.plot(
            x_values,
            np.angle(resp_pred, deg=True),
            color="tab:red",
            linewidth=3,
            linestyle=":",
        )
        ax_pz.scatter(
            np.real(zpk_pred.zeros),
            np.imag(zpk_pred.zeros),
            s=35,
            marker="o",
            ec="tab:red",
            fc="w",
            label="Fit Zeros",
        )
        ax_pz.scatter(
            np.real(zpk_pred.poles),
            np.imag(zpk_pred.poles),
            s=35,
            marker="x",
            ec="tab:red",
            fc="tab:blue",
            label="Fit Poles",
        )

    if xlim is not None:
        ax_amp.set_xlim(xlim)
        ax_phs.set_xlim(xlim)

    ax_amp.set_xscale("log")
    ax_amp.set_yscale("log")
    ax_amp.set_ylabel("Amplitude Response")
    ax_amp.grid()
    ax_amp.legend()

    ax_phs.set_ylim([-200.0, 200.0])
    ax_phs.set_xscale("log")
    ax_phs.set_ylabel("Phase Response")
    if x_units.lower() == "period":
        x_label = "Period (s)"
    elif x_units.lower() == "frequency":
        x_label = "Frequency (Hz)"
    ax_phs.set_xlabel(x_label)
    ax_phs.grid()

    ax_pz.set_xlabel("Re(z)")
    ax_pz.set_ylabel("Im(z)")
    max_lim = max(
        [
            abs(ax_pz.get_ylim()[0]),
            abs(ax_pz.get_ylim()[1]),
            abs(ax_pz.get_xlim()[0]),
            abs(ax_pz.get_xlim()[0]),
        ]
    )
    ax_pz.set_ylim([-1.25 * max_lim, 1.25 * max_lim])
    ax_pz.set_xlim([-1.25 * max_lim, 1.25 * max_lim])
    ax_pz.grid()
    ax_pz.legend()

    plt.show()


def plot_tf_obj(tf_obj, out_filename=None, show=True):
    """
    To Do: Get plotter from MTpy or elsewhere.
    See Issue #209
    https://github.com/simpeg/aurora/issues/209
    Parameters
    ----------
    tf_obj: aurora.transfer_function.TTFZ.TTFZ
        The transfer function values packed into an object
    out_filename: string
        Where to save the file.  No png is saved if this is False

    """
    from aurora.transfer_function.plot.rho_plot import RhoPlot
    import matplotlib.pyplot as plt

    plotter = RhoPlot(tf_obj)
    fig, axs = plt.subplots(nrows=2)
    ttl_str = tf_obj.tf_header.local_station.id
    plotter.rho_sub_plot(axs[0], ttl_str=ttl_str)
    plotter.phase_sub_plot(axs[1], ttl_str=ttl_str)
    if out_filename:
        plt.savefig(f"{out_filename}.png")
    if show:
        plt.show()



# -*- coding: utf-8 -*-
"""
"""

import numpy as np
import unittest

from aurora.time_series.apodization_window import ApodizationWindow


class TestApodizationWindow(unittest.TestCase):
    """
    Test ApodizationWindow
    """

    def setUp(self):
        pass

    #        self.band = Band()

    def test_default_boxcar(self):
        window = ApodizationWindow(num_samples_window=4)
        assert window.nenbw == 1.0
        assert window.coherent_gain == 1.0
        assert window.apodization_factor == 1.0
        # print(window.summary)

    def test_hamming(self):
        window = ApodizationWindow(taper_family="hamming", num_samples_window=128)
        assert np.isclose(window.nenbw, 1.362825788751716)
        assert np.isclose(window.coherent_gain, 0.54)
        assert np.isclose(window.apodization_factor, 0.6303967004989797)
        # print(window.summary)

    def test_blackmanharris(self):
        window = ApodizationWindow(
            taper_family="blackmanharris", num_samples_window=256
        )
        assert np.isclose(window.nenbw, 2.0043529382170493)
        assert np.isclose(window.coherent_gain, 0.35874999999999996)
        assert np.isclose(window.apodization_factor, 0.5079009302511663)
        # print(window.summary)

    def test_kaiser(self):
        apodization_window = ApodizationWindow(
            taper_family="kaiser",
            num_samples_window=128,
            taper_additional_args={"beta": 8},
        )
        print(apodization_window.summary)

    def test_tukey(self):
        apodization_window = ApodizationWindow(
            taper_family="tukey",
            num_samples_window=30000,
            taper_additional_args={"alpha": 0.25},
        )

        print(apodization_window.summary)

    def test_dpss(self):
        """ """
        apodization_window = ApodizationWindow(
            taper_family="dpss",
            num_samples_window=64,
            taper_additional_args={"NW": 3.0},
        )
        print(apodization_window.summary)

    def test_custom(self):
        apodization_window = ApodizationWindow(
            taper_family="custom",
            num_samples_window=64,
            taper=np.abs(np.random.randn(64)),
        )
        print(apodization_window.summary)
        print(apodization_window.summary)

        def test_initialization(self):
            for key in self.band.get_attribute_list():
                with self.subTest(key):
                    self.assertEqual(
                        self.band.get_attr_from_name(key),
                        self.band._attr_dict[key]["default"],
                    )


if __name__ == "__main__":
    # taw = TestApodizationWindow()
    # taw.test_blackmanharris()
    unittest.main()



# -*- coding: utf-8 -*-
"""
Extend DecimationLevel class with some aurora-specific methods
"""
# =============================================================================
# Imports
# =============================================================================
from mt_metadata.transfer_functions.processing.aurora import DecimationLevel


class DecimationLevel(DecimationLevel):
    def __init__(self, **kwargs):

        super().__init__(**kwargs)

    @property
    def windowing_scheme(self):
        from aurora.time_series.windowing_scheme import WindowingScheme

        windowing_scheme = WindowingScheme(
            taper_family=self.window.type,
            num_samples_window=self.window.num_samples,
            num_samples_overlap=self.window.overlap,
            taper_additional_args=self.window.additional_args,
            sample_rate=self.sample_rate_decimation,
        )
        return windowing_scheme

    # def to_stft_config_dict(self):
    #     """
    #     taper_family
    #     num_samples_window
    #     num_samples_overlap
    #     taper_additional_args
    #     sample_rate,
    #     prewhitening_type
    #     extra_pre_fft_detrend_type
    #     Returns
    #     -------
    #     """
    #     output = {}
    #     print(self.window.type)
    #     output["taper_family"] = self.window.type
    #     output["num_samples_window"] = self.window.num_samples
    #     output["num_samples_overlap"] = self.window.overlap
    #     output["taper_additional_args"] = self.window.additional_args
    #     output["sample_rate"] = self.sample_rate_decimation
    #     output["prewhitening_type"] = self.prewhitening_type
    #     output["extra_pre_fft_detrend_type"] = self.extra_pre_fft_detrend_type
    #     return output



import inspect
import os
import scipy.io as sio
import subprocess
import xarray as xr

from pathlib import Path

import aurora
import mt_metadata

init_file = inspect.getfile(aurora)
AURORA_PATH = Path(init_file).parent.parent
TEST_PATH = AURORA_PATH.joinpath("tests")
SANDBOX = AURORA_PATH.joinpath("aurora", "sandbox")
CONFIG_PATH = AURORA_PATH.joinpath("aurora", "config")
BAND_SETUP_PATH = CONFIG_PATH.joinpath("emtf_band_setup")
DATA_PATH = SANDBOX.joinpath("data")
DATA_PATH.mkdir(exist_ok=True, parents=True)
FIGURES_PATH = DATA_PATH.joinpath("figures")
FIGURES_PATH.mkdir(exist_ok=True, parents=True)
TEST_BAND_FILE = DATA_PATH.joinpath("bandtest.nc")
mt_metadata_init = inspect.getfile(mt_metadata)
MT_METADATA_DATA = Path(mt_metadata_init).parent.parent.joinpath("data")


def count_lines(file_name):
    """
    acts like wc -l in unix,
    raise FileNotFoundError: if file_name does not exist.

    Parameters
    ----------
    file_name: str or pathlib.Path
        The file to apply line counting to

    Returns
    -------
    num_lines: int
        Number of lines present in fileName or -1 if file does not exist

    """
    i = -1
    with open(file_name) as f:
        for i, l in enumerate(f):
            pass
    num_lines = i + 1
    return num_lines


def execute_subprocess(cmd, **kwargs):
    """

    Parameters
    ----------
    cmd : string
        command as it would be typed in a terminal
    kwargs

    Returns
    -------

    """
    """
    A wrapper for subprocess.call
    """
    exit_status = subprocess.call([cmd], shell=True, **kwargs)
    if exit_status != 0:
        raise Exception("Failed to execute \n {}".format(cmd))
    return


def execute_command(cmd, **kwargs):
    """
    Executes command in terminal from script.

    Parameters:
        cmd (str): command to exectute from a terminal
        kwargs: exec_dir (str): the directory from which to execute
        kwargs: no_exception: suppress output if exception

    Other Parameters:
        exit_status: :code:`0` is good, otherwise there is some problem

    .. note:: When executing :code:`rm *` this crashes if the directory we are removing
        from is empty

    .. note:: if you can you should probably use execute_subprocess() instead
    """
    exec_dir = kwargs.get("exec_dir", os.path.expanduser("~/"))
    allow_exception = kwargs.get("allow_exception", True)
    print("executing from {}".format(exec_dir))
    cwd = os.getcwd()
    os.chdir(exec_dir)
    exit_status = os.system(cmd)
    if exit_status != 0:
        print(f"exit_status of {cmd} = {exit_status}")
        if allow_exception:
            raise Exception(f"Failed to successfully execute \n {cmd}")
    os.chdir(cwd)


# <HDF5 save/load complex valued data>
def save_complex(data_array, *args, **kwargs):
    """
    netcdf and h5 do not handle complex values.  This method is a workaround.
    https://stackoverflow.com/questions/47162983/how-to-save-xarray-dataarray-with-complex128-data-to-netcdf
    Example Usage:
    band_da is an xarray
    save_complex(band_da, TEST_BAND_FILE)
    band_da = read_complex(TEST_BAND_FILE)

    Parameters
    ----------
    data_array
    args
    kwargs

    Returns
    -------

    """
    ds = xr.Dataset({"real": data_array.real, "imag": data_array.imag})
    return ds.to_netcdf(*args, **kwargs)


def read_complex(*args, **kwargs):
    ds = xr.open_dataset(*args, **kwargs)
    return ds["real"] + ds["imag"] * 1j


# </HDF5 save/load complex valued data>


def save_to_mat(data, variable_name, filename):
    """
    Example Usage:
    x = X.to_array(dim="channel")
    save_to_mat(x.data, "x", "x.mat")

    Reading into matlab or Octave:
    tmp = load("x.mat");
    data = tmp.x;

    Parameters
    ----------
    data : numpy array
        the data to save to file.  its fine if this is complex-valued.
    variable_name : string
        The name that we use to reference the variable within the struct in the matfile.
    filename : string
        The filepath to output

    Returns
    -------

    """
    sio.savemat(filename, {variable_name: data})
    return


class DotDict(dict):
    """
    Helper function for debugging, casts a dict so that its values
    can be accessed via dict.key as well as dict["key"]

    Usage:
    dot_dict = DotDict(basic_dict)
    """

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def __setattr__(self, name, value):
        self[name] = value



"""
See aurora issue #3.  This test confirms that the internal aurora stft
method returns the same array as scipy.signal.spectrogram
"""
import numpy as np

from aurora.pipelines.run_summary import RunSummary
from aurora.pipelines.time_series_helpers import prototype_decimate
from aurora.pipelines.time_series_helpers import run_ts_to_stft
from aurora.pipelines.time_series_helpers import run_ts_to_stft_scipy
from aurora.test_utils.synthetic.make_mth5_from_asc import create_test1_h5
from aurora.test_utils.synthetic.make_processing_configs import (
    create_test_run_config,
)
from aurora.transfer_function.kernel_dataset import KernelDataset
from mth5.mth5 import MTH5
from mth5.helpers import close_open_files


def test_stft_methods_agree():
    """
    The purpose of this method was to check if we could reasonably replace Gary's
    fft with scipy.signal.spectrogram.
    The answer is "mostly yes", under two conditons:
    1. scipy.signal.spectrogram does not inately support an extra linear detrending
    to be applied _after_ tapering.
    2. We do not wish to apply "per-segment" prewhitening as is done in some
    variations of EMTF.
    excluding this, we get numerically identical results, with basically
    zero-maintenance by using scipy.

    As of 30 Jun 2023, run_ts_to_stft_scipy is never actually used in aurora, except in
    this test.  That will change with the introduction of the FC layer in mth5 which
    will use that method.

    Because run_ts_to_stft_scipy will be used in mth5, we can port the aurora
    processing config to a mth5 FC processing config.  I.e. the dec_config argument to
    run_ts_to_stft can be reformatted so that it is an instance of
    mt_metadata.transfer_functions.processing.fourier_coefficients.decimation.Decimation

    """
    close_open_files()
    mth5_path = create_test1_h5()
    mth5_paths = [
        mth5_path,
    ]

    run_summary = RunSummary()
    run_summary.from_mth5s(mth5_paths)
    tfk_dataset = KernelDataset()
    station_id = "test1"
    run_id = "001"
    tfk_dataset.from_run_summary(run_summary, station_id)

    processing_config = create_test_run_config(station_id, tfk_dataset)

    mth5_obj = MTH5(file_version="0.1.0")
    mth5_obj.open_mth5(mth5_path, mode="a")

    for dec_level_id, dec_config in enumerate(processing_config.decimations):

        if dec_level_id == 0:
            run_obj = mth5_obj.get_run(station_id, run_id, survey=None)
            run_ts = run_obj.to_runts(start=None, end=None)
            local_run_xrts = run_ts.dataset
        else:
            local_run_xrts = prototype_decimate(dec_config.decimation, local_run_xrts)

        dec_config.extra_pre_fft_detrend_type = ""
        local_stft_obj = run_ts_to_stft(dec_config, local_run_xrts)
        local_stft_obj2 = run_ts_to_stft_scipy(dec_config, local_run_xrts)
        stft_difference = local_stft_obj - local_stft_obj2
        stft_difference = stft_difference.to_array()

        # drop dc term
        stft_difference = stft_difference.where(
            stft_difference.frequency > 0, drop=True
        )

        assert np.isclose(stft_difference, 0).all()

        print("stft aurora method agrees with scipy.signal.spectrogram")
    return


def main():
    test_stft_methods_agree()


if __name__ == "__main__":
    main()



"""
Placeholder module for methods manipulating xarray time series
"""

import xarray as xr


def handle_nan(X, Y, RR, drop_dim=""):
    """
    Drops Nan from multiple channel series'.
    Initial use case is for Fourier coefficients, but could be more general.

    Idea is to merge X,Y,RR together, and then call dropna.  We have to be careful
    with merging becuase there can be namespace clashes in the channel labels.
    Currently handling this by relabelling the remote reference channels from for
    example "hx"--> "remote_hx", "hy"-->"remote_hy".  If needed we could add "local" to
    local the other channels in X, Y.

    It would be nice to maintain an index of what was dropped.

    TODO: We can probably eliminate the config argument by replacing
    config.reference_channels with list(R.data_vars) and setting a variable
    input_channels to X.data_vars.  In general, this method could be robustified by
    renaming all the data_vars with a prefix, not just the reference channels

    Parameters
    ----------
    X : xr.Dataset
    Y : xr.Dataset or None
    RR : xr.Dataset or None
    drop_dim: string
        specifies the dimension on which dropna is happening.  For 3D STFT arrays
        this is "time", for 2D stacked STFT this is "observation"
    Returns
    -------
    X : xr.Dataset
    Y : xr.Dataset
    RR : xr.Dataset or None

    """
    if Y is None:
        Y = xr.Dataset()
    if RR is None:
        RR = xr.Dataset()

    input_channels = list(X.data_vars)
    output_channels = list(Y.data_vars)
    reference_channels = list(RR.data_vars)
    data_var_add_label_mapper = {}
    data_var_rm_label_mapper = {}
    for ch in reference_channels:
        data_var_add_label_mapper[ch] = f"remote_{ch}"
        data_var_rm_label_mapper[f"remote_{ch}"] = ch
    RR = RR.rename(data_var_add_label_mapper)

    merged_xr = X.merge(Y, join="exact")
    # Workaround for issue #228
    # merged_xr = merged_xr.merge(RR, join="exact")
    try:
        merged_xr = merged_xr.merge(RR, join="exact")
    except ValueError:
        print("Coordinate alignment mismatch -- see aurora issue #228 ")
        matches = X.time.values == RR.time.values
        print(f"{matches.sum()}/{len(matches)} timestamps match exactly")
        deltas = X.time.values - RR.time.values
        print(f"Maximum offset is {deltas.__abs__().max()}ns")
        #        print(f"X.time.[0]: {X.time[0].values}")
        #        print(f"RR.time.[0]: {RR.time[0].values}")
        merged_xr = merged_xr.merge(RR, join="left")
        for ch in list(RR.keys()):
            merged_xr[ch].values = RR[ch].values

    merged_xr = merged_xr.dropna(dim=drop_dim)
    merged_xr = merged_xr.to_array(dim="channel")
    X = merged_xr.sel(channel=input_channels)
    X = X.to_dataset(dim="channel")
    Y = merged_xr.sel(channel=output_channels)
    Y = Y.to_dataset(dim="channel")

    remote_channels = list(data_var_rm_label_mapper.keys())
    RR = merged_xr.sel(channel=remote_channels)
    RR = RR.to_dataset(dim="channel")
    RR = RR.rename(data_var_rm_label_mapper)

    return X, Y, RR


def cast_3d_stft_to_2d_observations(XY):
    """
    When the data for a frequency band are extracted from the STFT and
    passed to RegressionEstimator they have a typical STFT structure:
    One axis is time (the time of the window that was FFT-ed) and the
    other axis is frequency.  However we make no distinction between the
    harmonics (or bins) within a band.  We need to gather all the FCs for
        each channel into a 1D array.
    This method performs that reshaping (ravelling) operation.
    *It is not important how we unravel the FCs but it is important that
    we use the same scheme for X and Y.

    2021-08-25: Modified this method to use xarray's stack() method.

    Parameters
    ----------
    XY: either X or Y of the regression nomenclature.  Should be an
    xarray.Dataset already splitted on channel

    Returns
    -------
    output_array: numpy array of two dimensions (observations, channel)

    """
    if isinstance(XY, xr.Dataset):
        tmp = XY.to_array("channel")

    tmp = tmp.stack(observation=("frequency", "time"))
    return tmp
    # output_array = tmp.data.T
    # return output_array



from aurora.time_series.filters.filter_helpers import MT2SI_ELECTRIC_FIELD_FILTER


def triage_mt_units_electric_field(experiment):
    """
    One-off example of adding a filter to an mth5 in the case where the electric
    field data are given in V/m, but they were expected in mV/km.  This adds the
    correct filter to the metadata so that the calibrated data have units of
    mV/km.
     Parameters
    ----------
    experiment ;

    Returns
    -------

    """
    print(
        f"Add MT2SI_ELECTRIC_FIELD_FILTER to electric channels for parkfield here"
        f" {MT2SI_ELECTRIC_FIELD_FILTER} "
    )
    filter_name = MT2SI_ELECTRIC_FIELD_FILTER.name
    survey = experiment.surveys[0]
    survey.filters[filter_name] = MT2SI_ELECTRIC_FIELD_FILTER
    stations = survey.stations
    for station in stations:
        channels = station.runs[0].channels
        for channel in channels:
            if channel.component[0] == "e":
                channel.filter.name.insert(0, filter_name)
                channel.filter.applied.insert(0, False)
    return experiment


def triage_missing_coil_hollister(experiment):
    """
    One off for hollister missing hy metadata for no reason I can tell
    Parameters
    ----------
    experiment

    Returns
    -------

    """
    survey = experiment.surveys[0]
    stations = survey.stations
    for station in stations:
        if station.id == "SAO":
            runs = station.runs
            for run in runs:
                channels = run.channels
                print(channels)
                for channel in channels:
                    print(channel.id)

        # station = stations[i_station]
        runs = station.runs[0]
        print("help")



"""
An extension of RegressionEstimator with properties common to Gary's TRME and TRME_RR
see Notes in RME, RME_RR
"""
import numpy as np
import xarray as xr

from copy import deepcopy
from scipy.linalg import solve_triangular

from aurora.transfer_function.regression.base import RegressionEstimator
from aurora.transfer_function.regression.helper_functions import rme_beta


class MEstimator(RegressionEstimator):
    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs
        expectation_psi_prime : numpy array
            Same number of entries as there are output channels (columns of Y)
            Default is a vector of 1.0's

            Expectation value of psi' (derivative of psi) -- ?? OR rho'=psi????
            Recall start with the loss function rho. The derivative of the
            loss function is the "influence function" psi.
            Think about the Huber loss function (quadratic out to r0,
            after which it is linear). Psi is -1 until you get to -r0,
            then it increases linearly to 1 (r0)
            Psi' is something like 1 between -r0, and r0
            Psi' is zero outside
            So the expectiaon value of psi' is the number of points outside
            its the number of points that didnt get weighted /total number of points

        """
        super(MEstimator, self).__init__(**kwargs)
        self.expectation_psi_prime = np.ones(self.n_channels_out)
        self.Yc = deepcopy(self.Y)
        self._Y_hat = None
        self._residual_variance = None

    @property
    def QHYc(self):
        if self._QHYc is None:
            self.update_QHYc()
        return self._QHYc

    def update_QHYc(self):
        self._QHYc = self.QH @ self.Yc

    @property
    def Y_hat(self):
        if self._Y_hat is None:
            self.update_y_hat()
        return self._Y_hat

    def update_y_hat(self):
        print("Y_hat update method is not defined for abstract MEstimator class")
        print("Try using RME or RME_RR class instead")
        raise Exception

    def update_residual_variance(self, correction_factor=1):
        print("update_residual_variance method not defined in abstract MEstimator")
        print("Try using RME or RME_RR class instead")
        raise Exception

    @property
    def residual_variance(self):
        if self._residual_variance is None:
            self.update_residual_variance()
        return self._residual_variance

    @property
    def r0(self):
        return self.iter_control.r0

    @property
    def u0(self):
        return self.iter_control.u0

    # @property
    # def beta(self):
    #     return rme_beta(self.r0)

    @property
    def correction_factor(self):
        """
        *May want to move this out of iter_control, it is really only controlled by r0.

        Returns
        -------
        correction_factor: float
            See doc in iter_control.IterControl.correction_factor()
        """
        return self.iter_control.correction_factor

    def residual_variance_method1(self):
        """
        This is the method that was originally in TRME_RR.m.  It seems more correct
        than the one in TRME, but also has more computational overhead.
        """
        res = self.Yc - self.Y_hat  # intial estimate of error variance
        residual_variance = np.sum(np.abs(res * np.conj(res)), axis=0) / self.n_data
        return residual_variance

    def residual_variance_method2(self):
        """
        These are the error variances.

        Computes the squared norms difference of the output channels from the
        "output channels inner-product with QQH"

        This method used to take either QHY, Y as arguments or QHYc, Yc
        But now that Yc initializes to Y, we can just use with QHYc, Yc always
        (Note QHYc is updated everytime Yc is updated)

        Parameters
        ----------
        QHY : numpy array
            QHY[i,j] = Q.H * Y[i,j] = <Q[:,i], Y[:,j]>
            So when we sum columns of norm(QHY) we are get in the zeroth position
            <Q[:,0], Y[:,0]> +  <Q[:,1], Y[:,0]>, that is the 0th channel of Y
            projected onto each of the Q-basis vectors
        Y_or_Yc : numpy array
            The output channels (self.Y) or the cleaned output channels self.Yc
        correction_factor : float
            See doc in IterControl.correction_factor

        Returns
        -------
        residual_variance : numpy array
            One entry per output channel.

        """
        Y2 = np.linalg.norm(self.Yc, axis=0) ** 2  # variance?
        QHY2 = np.linalg.norm(self.QHYc, axis=0) ** 2
        residual_variance = (Y2 - QHY2) / self.n_data

        try:
            assert (residual_variance > 0).all()
        except AssertionError:
            print("WARNING - Negative error variances observed")
            print(residual_variance)
            print("Setting residual_variance to zero - Negative values observed")
            residual_variance *= 0

        return residual_variance

    def update_y_cleaned_via_huber_weights(self):
        """
        Updates the values of self.Yc and self.expectation_psi_prime

        Parameters
        ----------
        residual_variance : numpy array
            1D array, the same length as the number of output channels
            see self.residual_variance() method for its calculation
        Y_hat : numpy array
            The predicted data, usually from QQHY

        Returns
        -------
        None

        Original matlab documenation:
        function [YC,E_psiPrime] = HuberWt(Y,YP,sig,r0)
        inputs are data (Y) and predicted (YP), estiamted
        error variances (for each column) and Huber parameter r0
        allows for multiple columns of data
        """
        for k in range(self.n_channels_out):
            r0s = self.r0 * np.sqrt(self.residual_variance[k])
            residuals = np.abs(self.Y[:, k] - self.Y_hat[:, k])
            w = np.minimum(r0s / residuals, 1.0)
            self.Yc[:, k] = w * self.Y[:, k] + (1 - w) * self.Y_hat[:, k]
            self.expectation_psi_prime[k] = 1.0 * np.sum(w == 1) / self.n_data
        self.update_QHYc()  # note the QH is different in TRME_RR vs TRME
        return

    def initial_estimate(self):
        """
        Make first estimate of TF (b), Y_hat, and residual_variance
        """
        self.qr_decomposition()
        self.update_b()
        self.update_y_hat()
        self.update_residual_variance()

    def apply_huber_regression(self):
        """This is the 'convergence loop' from TRME, TRME_RR"""
        converged = self.iter_control.max_number_of_iterations <= 0
        self.iter_control.number_of_iterations = 0
        while not converged:
            b0 = self.b
            self.iter_control.number_of_iterations += 1
            self.update_y_cleaned_via_huber_weights()
            self.update_b()
            self.update_y_hat()
            self.update_residual_variance(correction_factor=self.correction_factor)
            converged = self.iter_control.converged(self.b, b0)
        return

    def apply_redecending_influence_function(self):
        """one or two iterations with redescending influence curve cleaned data"""
        if self.iter_control.max_number_of_redescending_iterations:
            self.iter_control.number_of_redescending_iterations = 0  # reset per channel
            while self.iter_control.continue_redescending:
                self.iter_control.number_of_redescending_iterations += 1
                self.update_y_cleaned_via_redescend_weights()
                self.update_b()
                self.update_y_hat()
                self.update_residual_variance()
            # crude estimate of expectation of psi
            # accounts for redescending influence curve
            self.expectation_psi_prime = 2 * self.expectation_psi_prime - 1
        return

    def estimate(self):
        """
        function that does the actual remote reference estimate

        Here is a comment from the matlab codes:
        "need to look at how we should compute adjusted residual cov to make
         consistent with tranmt"
        See issue#69 aurora github repo addresses this
        """
        if self.is_underdetermined:
            self.solve_underdetermined()
            return

        self.initial_estimate()
        self.apply_huber_regression()
        self.apply_redecending_influence_function()

        if self.iter_control.return_covariance:
            self.compute_inverse_signal_covariance()
            self.compute_noise_covariance()
            self.compute_squared_coherence()
        return

    def update_y_cleaned_via_redescend_weights(self):
        """
        Updates estimate for self.Yc as a match-filtered sum of Y and Y_hat.


        Parameters
        ----------
        Y_hat: numpy array
            An estimate of the output data Y obtained by
            self.Q @ self.QHY
            or
            self.Q @ self.QHYc
        residual_variance

        Returns
        -------

        Matlab documentation:
        function[YC, E_psiPrime] = RedescendWt(Y, YP, sig, u0):
        inputs are data(Y) and predicted(YP), estimated error variances (for each
        column) and Huber parameter u0.  Allows for multiple columns of data
        """
        for k in range(self.n_channels_out):
            sigma = np.sqrt(self.residual_variance[k])
            r = np.abs(self.Y[:, k] - self.Y_hat[:, k]) / sigma
            t = -np.exp(self.u0 * (r - self.u0))
            w = np.exp(t)

            # cleaned data
            self.Yc[:, k] = w * self.Y[:, k] + (1 - w) * self.Y_hat[:, k]

            # computation of E(psi')
            t = self.u0 * (t * r)
            t = w * (1 + t)
            self.expectation_psi_prime[k] = np.sum(t[t > 0]) / self.n_data
        self.update_QHYc()
        return

    def compute_squared_coherence(self):
        """
        res: Residuals, the original data minus the predicted data.
        SSR : Sum of squares of the residuals, per channel

        """
        res = self.Y - self.Y_hat
        SSR = np.einsum("ij,ji->i", res.conj().T, res)
        SSR = np.real(SSR).T
        Yc2 = np.abs(self.Yc) ** 2
        SSYC = np.sum(Yc2, axis=0)
        R2 = 1 - SSR / SSYC
        R2[R2 < 0] = 0

        self.R2 = xr.DataArray(
            R2,
            dims=[
                "output_channel",
            ],
            coords={
                "output_channel": list(self._Y.data_vars),
            },
        )

        return

    def compute_noise_covariance(self):
        """
        res_clean: The cleaned data minus the predicted data. The residuals
        SSR_clean: Sum of squares of the residuals.  Diagonal is real
        Parameters
        ----------
        Y_hat

        Returns
        -------

        """
        res_clean = self.Yc - self.Y_hat
        SSR_clean = np.conj(res_clean.conj().T @ res_clean)
        inv_psi_prime2 = np.diag(1.0 / (self.expectation_psi_prime ** 2))
        cov_nn = inv_psi_prime2 @ SSR_clean / self.degrees_of_freedom

        self.cov_nn = xr.DataArray(
            cov_nn,
            dims=["output_channel_1", "output_channel_2"],
            coords={
                "output_channel_1": list(self._Y.data_vars),
                "output_channel_2": list(self._Y.data_vars),
            },
        )
        return



"""
There are several ways to go about this.  One reasonably straightforward one is:

Take all the Fourier coefficients right before the go into regression (right before or
after we apply the effective degrees of freedom weights).  Data are strictly 2D at
this point, you have each channel of data corresponding to a column, and each row
to an observation.
Following Jones and Jodicke, lets take the coherence of Ex and Hy
and that of Ey and Hx.  This yields one single number.

Then, iteratively remove each observation in turn and count whether that partial
cohernece is bigger or smaller than the baseline and by how much

Deflection downward means naughty data and defelction  upward means well-behaved data.

This leads to a coherence score for each obervation.  We can boot out all data that
 say move the estimate down (worse than average)
"""

import numpy as np


def drop_column(x, i_col):
    """

    Parameters
    ----------
    x
    i_drop

    Returns
    -------

    """
    return np.hstack([x[:, 0:i_col], x[:, i_col + 1 : :]])


def coherence_from_fc_series(c_xy, c_xx, c_yy):
    n_obs = len(c_xy)
    num = np.sum(c_xy) ** 2
    den = np.sum(c_xx * c_yy)
    coh = (num / den) / n_obs
    return coh


def coherence_weights_v00(x, y, threshold=0.95):  # 975):#0.98
    """

    Parameters
    ----------
    x: X["hx"].data
    y: Y["ey"].data
    threshold : value in [0,1] to set what to keep / reject

    Returns
    -------
    W: Weights (currently set to be 0 or 1)
    """
    # Initialize a weight vector the length = num_observations
    n_obs = len(x)
    partial_coh = np.zeros(n_obs)
    W = np.zeros(n_obs)  # for example

    c_xy = np.abs(x * np.conj(y))
    c_xx = np.real(x * np.conj(x))
    c_yy = np.real(y * np.conj(y))

    ccc = np.vstack([c_xy, c_xx, c_yy])

    for i in range(n_obs):
        partial_series = drop_column(ccc, i)
        partial_coh[i] = coherence_from_fc_series(
            partial_series[0, :],
            partial_series[1, :],
            partial_series[2, :],
        )

    worst_to_best = np.argsort(partial_coh)
    # coh0 = coherence_from_fc_series(c_xy, c_xx, c_yy)
    # relative_coherence = partial_coh / coh0
    # sorted_partials = partial_coh[worst_to_best]
    clip_point = int(threshold * n_obs)
    keepers = worst_to_best[clip_point:]
    W[keepers] = 1
    return W


def compute_coherence_weights(X, Y, RR, coh_type="local"):
    """
    2022-09-09: This method is not yet supported.  It needs to be made
    tolerant of channel_nomenclature.

    Parameters
    ----------
    X
    Y
    RR
    coh_type: "local" or "remote"

    Returns
    -------

    """
    remote_threshold = 0.8
    local_threshold = 0.95

    X = X.dropna(dim="observation")
    Y = Y.dropna(dim="observation")
    if RR is not None:
        RR = RR.dropna(dim="observation")

    # < INSERT COHERENCE SORTING HERE>    y_type = "remote"
    null_indices = X["hx"].isnull()  # not robust -- hail mary
    finite_indices = ~null_indices
    W = np.zeros(len(X.observation))

    x = X["hx"]  # .dropna(dim="observation").data

    if coh_type == "local":
        y = Y["ey"]  # .dropna(dim="observation").data
        threshold = local_threshold
    elif coh_type == "remote":
        y = RR["hx"]  # .dropna(dim="observation").data
        threshold = remote_threshold

    W1 = coherence_weights_v00(x, y, threshold=threshold)

    W[finite_indices] = W1

    W[W == 0] = np.nan
    X["hx"].data *= W
    Y["ey"].data *= W

    # x = X["hy"].data
    # y = Y["ex"].data
    W = np.zeros(len(finite_indices))
    x = X["hy"]  # .dropna(dim="observation").data
    if coh_type == "local":
        y = Y["ex"]  # .dropna(dim="observation").data
        threshold = local_threshold
    elif coh_type == "remote":
        y = RR["hy"]  # .dropna(dim="observation").data
        threshold = remote_threshold
    W2 = coherence_weights_v00(x, y)
    W[finite_indices] = W2
    W[W == 0] = np.nan
    X["hy"].data *= W
    Y["ex"].data *= W
    # W = W*W2
    # X *= W
    # Y *= W
    if RR is not None:
        RR *= W
    return X, Y, RR
